from sner.scripts import context, names, spelling, utilities
import codecs
import csv


def main(data, options):
    """
	Currently this region of code is not in full working order.  It originally
	analyzed the corpus to look for potential rules to identify names.
	Specifically it would generate every possible spelling and context rule,
	and then assess its performance.

    Args:
        data(string) = the name of the .csv file of the Garshana corpus.  (The one
			   containing the full text of various tablets.
        options = object containing various configuration information (i believe
				  this type no longer exists in the code base, it was intended
				  to be replaced with the config object)

    Returns:
        None

    Raises:
        None

    """

    file = codecs.open('data/Attestations_PNs.csv','r', encoding='utf-16')

    for line in file:
        line = line.split(',')
        if line[9].rstrip() == 'PN':
            text = utilities.clean_line(line[4].rstrip())
            name = utilities.clean_line(line[5].rstrip())
            if name in names.personal:
                names.personal[name] += 1
            else:
                names.personal[name] = 1
            context.main(text, name)

    #get a dictionary of words to their occurences, and syllables to their
    #occurences respectively
    word_count, syll_count = utilities.get_counts(data)
    rules_collection = [
            [context.left_rules,  word_count, 'Left'],
            [context.right_rules, word_count, 'Right']]

    # Do spelling rule analysis
    spelling.main(data, syll_count)

	# do context rule analysis
    with open('results/context.csv', 'w', newline='',
              encoding='utf-8') as csvfile:
        fieldnames = [
                'Context',
                'Rule',
                'Occurrence',
                'Total Occurrence',
                'Percentage']
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()

        for rules in rules_collection:
            for rule in rules[0]:
                if rule in rules[1]:
                    total = rules[1][rule]
                    percent = float(rules[0][rule]) / float(total)
                    writer.writerow({
                        'Context': rules[2],
                        'Rule': rule,
                        'Occurrence': str(rules[0][rule]),
                        'Total Occurrence': str(total),
                        'Percentage': str(percent)})

    with open('results/varients.csv', 'w', newline='',
              encoding='utf-8') as csvfile:
        fieldnames = [
                'Name',
                'Varients']
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()

        for name in context.varients.keys():
            writer.writerow({
                'Name': name,
                'Varients': str(context.varients[name])})

    with open('results/namesperrule.csv', 'w', newline='',
              encoding='utf-8') as csvfile:
        fieldnames = ['Rule', 'Name', 'Occurrences']
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()
        for rule in context.rules:
            for name in context.rules[rule]:
                writer.writerow({
                    'Rule': rule,
                    'Name': name,
                    'Occurrences': str(context.rules[rule][name])})

if __name__ == '__main__':
    main()



"""Display Class"""
import time

class Display:
    """Display the progress of a process.
	
	Will print a progress bar updated by function calls made off of an
	instance of this type.  Will also estimate time to completion via naieve
	averaging.

    Attributes:
        start_time (float): Seconds since epoch to when progress starts.
        elapsed_time (float): Seconds since progress started.
        last_updated (float): Seconds since epoch to when progress was
            last updated.
    """

    def __init__(self):
        self.start_time = None
        self.elapsed_time = None
        self.last_updated = None

    def start(self, message=None):
        """Used to denote the beginning of whatever process you are displaying
		the progress of.  Will mark down the current time for purposes of
		estimating the time remaining.  May display an initial message that
		you can pass in.
		
        Args:
            message (string): Optional start message.

        Returns:
            None

        Raises:
            None
        """

        self.start_time = time.time()
        if message:
            print(message)

    def update_progress_bar(self, step, end):
        """Call at the end of each 'iteration', whatever that happens to mean
		in the current context.  Will update the progress bar and estimated
		time remaining accordingly.
		
        Args:
            step (float): Current iteration of process.
            end (float): Final iteration of process.

        Returns:
            None

        Raises:
            None
        """

        percent = float(step) / float(end)

        start_time = self.start_time
        current_time = time.time()

        if self.last_updated and current_time < self.last_updated + 0.017:
            return
        else:
            self.last_updated = current_time

        elapsed_time = current_time - start_time
        self.elapsed_time = elapsed_time
        estimated_time = (elapsed_time / percent) - elapsed_time
        hours = int(estimated_time / 3600.0)
        minutes = int((estimated_time - (hours * 3600)) / 60.0)
        seconds = int(estimated_time - (minutes * 60) - (hours * 3600))
        time_remaining = "{:02d}:{:02d}:{:02d}".format(
            hours,
            minutes,
            seconds
        )

        progress_bar = "{}".format('\u2588' * int(percent * 25.0))
        remainder = (percent * 25.0) - len(progress_bar)

        if remainder >= 0.75:
            progress_bar += '\u258a'
        elif remainder >= 0.5:
            progress_bar += '\u258c'
        elif remainder >= 0.25:
            progress_bar += '\u258e'

        progress_bar += ' ' * (25 - len(progress_bar))
        output = "    {:05.2f}% |{}| Time Remaining: {}".format(
            percent * 100.0,
            progress_bar,
            time_remaining
        )
        print(' ' * 80, end='\r')
        print(output, end='\r')

    def finish(self):
        """Displays elapsed time of the tracked process. Clears attributes.
		Call upon the completion of whatever it is you are tracking.

        Args:
            None

        Returns:
            None

        Raises:
            None
        """

        hours = int(self.elapsed_time / 3600.0)
        minutes = int(self.elapsed_time / 60.0)
        seconds = int(self.elapsed_time - (minutes * 60) - (hours * 3600))
        elapsed_time = "{:02d}:{:02d}:{:02d}".format(
            hours,
            minutes,
            seconds
        )

        print("   100.00% |{}|   Elapsed Time: {} ".format(
            '\u2588' * 25,
            elapsed_time
        ))

        self.start_time = None
        self.elapsed_time = None
        self.last_updated = None



# Will assign a strength rating to the rule provided as a paramter.
# ProspectiveNames refers to the tokens this rule identified as names.
# It will check the annotations of these tokens to see if they are really names
# then it will use that to compute a strength rating

from sner.classes import Rule, Token
from sner.scripts.ner import namesfromrule


# Rate the strength of the individual rule passed in,
# given all Tokens where the rule is true (identifiedNames)
def rateRulePerformance(results, rule, alpha, k, accept_threshold):
    total = len(results)
    name_count = 0
    for token in results:
        if token.name_probability > accept_threshold:
            name_count += 1

    strength = (name_count + alpha) / (total + k * alpha)

    rule.strength = strength
    rule.occurrences = total


# Find all tokens that match a given rule
# This is used to rate a rules performance, by totalling up the
# tokens that match the rule,
# and comparing that to how many of the tokens in that set are marked as a PN


def main(corpus, rules, options, iteration, display):
    """
    Finds all tokens that match a given rule, using that to rate the rules
    performance. Rates rule performance by totalling up the tokens that
    match the rule, and comparing that to how many of the tokens in that set
    are considered to be a PN.
    Args:
        corpus = Set of all lines from the corpus.
        rules = RuleSet object of all currently used rules.
        options = values pulled from the configuration file.

    Returns:
        None

    Raises:
        None

    """

    alpha = options.alpha
    k = options.k
    accept_threshold = options.accept_threshold

    i = 0
    for rule in rules:
        names = namesfromrule.main(corpus, rule)
        rateRulePerformance(names, rule, alpha, k, accept_threshold)

        i += 1
        display.update_progress_bar(
            (len(rules) * (iteration - 1)) + i,
            len(rules) * options.iterations
        )



"""Object representing the various types of names.  Formatted as dictionaries
mapping plaintext versions of names to the number of occurrences of that names
"""

#!/usr/bin/python3

# Mike Canoy
# Authored 5/9/2016

personal = {}

geographic = {}

field = {}



"""Token"""
from enum import Enum


class Token:
    """
    This represents a given 'token' in the corpus.
    It represents a combination of left context, right context, and the word itself.
    Any identical tokens will be combined into pre-existing ones.
	
	Attributes:
        left_context (str): contents of token to the left of this token
        word (str): this token as a string
        right_context (str): contents of token to the right of this token
		type (Enum): what kind of data does this token represent? (personal name, date, etc)
		rules (set): set of Rule objects that apply to this token
        name_probability (float): percent chance that this token represents a name
    """

    Type = Enum('Types', 'none personal_name geographic_name profession number date')

    def __init__(self, left_context, word, right_context, token_type):
        self.left_context = left_context
        self.word = str(word)
        self.right_context = right_context
        self.type = token_type
        self.rules = set()
        self.name_probability = 0.0

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return self.word == other.word

    def __nq__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "Token(left_context={}, word={}, right_context={}, {})".format(
            self.left_context,
            self.word,
            self.right_context,
            self.type
        )

    @classmethod
    def find_type(cls, token_type):
        """Parse a string representation and return an equivalent Token.Type
		enumeration value.  A string representation is used in the input data
		files, however the enumeration allows for faster comparisons.  This
		function is a convenience to make it easier to scan the appropriate
		input files.
		
        Args:
            token_type (string): String representation of a type of token.

        Returns:
            type (Token.Type): Enumeration value representing the type of token.

        Raises:
            TypeError: If the string represents some unrecognized type of
					   token, then freak out.  You should probably respond to 
					   this by adding said rule to the enumeration above as 
					   well as to the code here.
        """

        if token_type == 'PN':
            return Token.Type.personal_name

        elif token_type == 'GN':
            return Token.Type.geographic_name

        elif token_type == 'PF':
            return Token.Type.profession

        elif token_type == 'N':
            return Token.Type.number

        elif token_type == 'D':
            return Token.Type.date

        elif token_type == '-':
            return Token.Type.none

        else:
            raise TypeError("Unrecognized Word Type: " + word_type)



""" Rule """
from enum import Enum


class Rule:
    """Object representing a rule that is used to identify names within the 
	corpus.  This object is used in the ner model, rather than the sner model.

    Attributes:
        type (Enum): What variety of rule is this (spelling, context, etc)
        contents (str): Data that defines the rule
        strength (float): Percent chance that any token this rule applies to 
					      is a name
        occurrences (int): Number of tokens this rule applies to
		iteration (int): The iteration in which this rule was generated
    """

	#enumeration representing the various rule types, these are used to speed
	#comparisons between rules
    Type = Enum('Types', 'unset spelling left_context right_context')

    def __init__(self, ruletype, rule, strength):
        self.type = ruletype
        self.contents = str(rule)
        self.strength = float(strength)
        self.occurrences = 1
        self.iteration = -1

    def __eq__(self, other):
        return self.key() == other.key()

    def __gt__(self, other):
        return self.strength > other.strength

    def __nq__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.key())

    def __lt__(self, other):
        return self.strength < other.strength

    def __str__(self):
        return "Rule(type={}, rule={}, strength={})".format(
            self.type.name,
            self.contents,
            self.strength
        )

    def key(self):
        """Define what constitutes an identifying key for a Rule type object.
		The Rule's type as a string concatenated with the Rule's contents was
		considered to be uniquely identifying.
		
        Args:
            None

        Returns:
            key (string): Uniquely identifying key for this Rule object.

        Raises:
            None
        """

        return "{}:{}".format(self.type.value, self.contents)

    @classmethod
    def find_type(cls, rule_type):
        """Parse a string representation and return an equivalent Rule.Type
		enumeration value.  A string representation is used in the input data
		files, however the enumeration allows for faster comparisons.  This
		function is a convenience to make it easier to scan the appropriate
		input files.
		
        Args:
            rule_type (string): String representation of a type of rule.

        Returns:
            type (Rule.Type): Enumeration value representing the type of rule.

        Raises:
            TypeError: If the string represents some unrecognized type of rule,
					   then freak out.  You should probably respond to this by
					   adding said rule to the enumeration above as well as to
					   the code here.
        """

        if rule_type == 'LEFT_CONTEXT':
            return Rule.Type.left_context
        elif rule_type == 'RIGHT_CONTEXT':
            return Rule.Type.right_context
        elif rule_type == 'SPELLING':
            return Rule.Type.spelling
        else:
            raise TypeError



from sner.classes import Rule, Token


def main(ruleset, maxrules):
    """
    Determine which rules will be accepted by an iteration of the algorithm

    Sorts rules into the top n rules, where n = maxrules
    Sorted by strength, ties broken by alphabetization

    Args:
        ruleset = set of all rules known
        maxrules = integer value of maximum number of rules we can accept
    Returns:
        The set of rules that are to be accepted by the next iteration.

    Raises:
        None

    """

    #sort the rules from best to worst (uses the __lt__ function defined in Rule class (in ruleset.py))
    sortedlist = sorted(ruleset, reverse=True)

    #trim the list down to only the rules that will go on to the next iteration
    sortedlist = sortedlist[:maxrules]

    rules = set()

    for rule in sortedlist:
        rules.add(rule)

    return rules



from sner.scripts import utilities
from collections import defaultdict
import codecs
import editdistance


def getPersonalNames(csvFile):
    """
    Args:
        csvfile = A csv file with the name on line[5] and line[9] is an identifier of its
                  type, where we are looking for 'PN' indicating a personal name.


    Returns:
        Puts found names from the csv file and returns them in a dictionary
         of the form: { Name : Occurrences }

    Raises:
        None

    """

    names = defaultdict(int)
    try:
        open_file = codecs.open(csvFile, 'r', encoding='utf-16')
        for line in open_file:
            line_list = line.split(',')
            name = line_list[5].rstrip()
            name = utilities.clean_line(name)

            text = line_list[4].rstrip()
            text = utilities.clean_line(text)
            text = text.split(' ')

            edit_distance = []
            for word in text:
                edit_distance.append(editdistance.eval(name, word))

            name = text[edit_distance.index(min(edit_distance))]
            nameType = line_list[9].rstrip()
            if (nameType == "PN"):
                names[name] = names[name] + 1
        open_file.close()
    except IOError:
        print("Cannot open: " + csvFile)

    return names


def getPNs(data):
    """
    Collects names in Garshana csv and returns them
    """
    return getPersonalNames(data.attestations)


def getKgrams(names, k):
    """
    Args:
        names = dictionary of the form { Names : Occurrences }
        k = Max k-gram you want to retrieve.
            Special values of k = -1 returning dictionary of monograms.
                          and k = -2 returning dictionary of monograms,
                                  followed by a dictionary of bigrams.

        Example use to get all monograms, bigrams, and trigrams:
        getKgrams(getPNs(), 3)


    Returns:
        A dictionary of all grams up to order k

    Raises:
        None

    """

    grams = [None] * k

    for i in range(0, k):
            grams[i] = defaultdict(int)

    for name, value in names.items():
        syllables = name.split('-')

        for i in range(1, k + 1):
            end = len(syllables) - i + 1
            for j in range(0, end):
                gram = '-'.join(syllables[j:j + i])
                grams[i-1][gram] = grams[i-1][gram] + value

    return grams


def testKgrams():
    """
    A simple test to print off Trigrams and Quadgrams that
    occur more than once [note: duplicate names will trigger these]
    """
    nameDict = getPNs(data)
    # sampleNames = {"ab-ac-ad-ae" : 3, "ab-ae-ad" : 5, "ae-ac-ac-ad" : 1}
    testGrams = getKgrams(nameDict, 4)
    for i in range(2, 4):
        for key, value in testGrams[i].iteritems():
            if (value > 1):
                print("Gram: " + key + " seen : " + str(value))



"""
Sumerian Named Entity Recognition: sner.py

This file bootstraps the program, parsing arguments and configuration file.
Also offers interface for our analysis and formatting tools and test functions.
"""
import argparse
import json
import os
#import pytest
from sner.models import ner, sklearn_launcher
from sner.scripts import export, export_atf, overfit_check, output # analysis, export, formatting


def bool_check(arg, config, default):
    """Return bool from `arg` or `config` str, or `default` bool.

    Arguments from the parser and configuration file are represented as strings.
    Arguments that represent boolean values need string compares, and then
    return the actual boolean object.

    Args:
        arg (str):
        config (str):
        default (bool):

    Returns:
        True, False, or default (bool):

    Raises:
        TypeError
    """

    if arg:
        if arg == 'True':
            return True
        elif arg == 'False':
            return False
        else:
            raise TypeError
    elif config:
        if config == 'True':
            return True
        elif config == 'False':
            return False
        else:
            raise TypeError
    else:
        return default



def add_args(parser):
    """
    Adds arguments from the command line and interprets them as appropriate.
    Available arguments:

    Args:
        parser

    Returns:
        None

    Raises:
        None

    """

    parser.add_argument(
        '-r',
        '--run',
        help='Run one of the following models: '
        '[dec], '
        '[nbc], '
        '[ner], '
        '[rdf], '
        '[sgd], '
        '[forest], '
        'or [svc]. '
        'Or one of the following routines: '
        '[analysis], '
        '[export], '
        '[export-atf], '
        '[formatting], '
        '[over-fit], '
        '[output], '
        'or [testing].',
        required=False,
        choices=[
            'dec',
            'nbc',
            'ner',
            'rdf',
            'sgd',
            'svc',
            'analysis',
            'export',
            'export-atf',
            'formatting',
            'testing',
            'over-fit',
            'forest',
            'output'
        ]
    )
    parser.add_argument(
        '-cf',
        '--config',
        help='Configuration file to use',
        required=False
    )
    parser.add_argument(
        '-p',
        '--path',
        help='Path to data directory',
        required=False
    )
    parser.add_argument(
        '-c',
        '--corpus',
        help='File name of the corpus',
        required=False
    )
    parser.add_argument(
        '-a',
        '--attestations',
        help='File name of the attestations',
        required=False
    )
    parser.add_argument(
        '-sr',
        '--seed-rules',
        help='File name of the seed rules',
        required=False
    )
    parser.add_argument(
        '-i',
        '--iterations',
        help='Number of iterations',
        type=int,
        required=False
    )
    parser.add_argument(
        '-mr',
        '--max-rules',
        help='Max number of rules per iterations',
        type=int,
        required=False
    )
    parser.add_argument(
        '-al',
        '--alpha',
        help='Alpha value',
        type=float,
        required=False
    )
    parser.add_argument(
        '-k',
        '--k',
        help='K value',
        type=float,
        required=False
    )
    parser.add_argument(
        '-nd',
        '--norm-date',
        help='Enable date normalization',
        required=False,
        choices=['True', 'False']
    )
    parser.add_argument(
        '-ng',
        '--norm-geo',
        help='Enable geographical name normalization',
        required=False,
        choices=['True', 'False']
    )
    parser.add_argument(
        '-nn',
        '--norm-num',
        help='Enable number normalization',
        required=False,
        choices=['True', 'False']
    )
    parser.add_argument(
        '-np',
        '--norm-prof',
        help='Enable profession normalization',
        required=False,
        choices=['True', 'False']
    )
    parser.add_argument(
        '-dn',
        '--disjoint-names',
        help='Export train / dev with disjoint unique name sets',
        required=False,
        choices=['True', 'False']
    )
    parser.add_argument(
        '-el',
        '--export-left-rules',
        help='Export uses hardcoded left context rules',
        required=False,
        choices=['True', 'False']
    )
    parser.add_argument(
        '-er',
        '--export-right-rules',
        help='Export uses hardcoded right context rules',
        required=False,
        choices=['True', 'False']
    )
    parser.add_argument(
        '-es',
        '--export-spelling-rules',
        help='Export uses hardcoded spelling rules',

        required=False,
        choices=['True', 'False']
    )
    parser.add_argument(
        '-ef',
        '--export-found',
        help='Export creates features out of found syllables',
        required=False,
        choices=['True', 'False']
        
    )
    parser.add_argument(
        '-atf',
        '--use-atf',
        help='Enables use of the atf files for the scikit models',

        required=False,
        choices=['True', 'False']
    )
    parser.add_argument(
        '-ds',
        '--dev-size',
        help='Percentage for exporter to use for dev set. 0 > size >= 1',
        type=float,
        required=False
    )
    parser.add_argument(
        '-ts',
        '--train-size',
        help='Percentage for exporter to use for train set. 0 > size >= 1',
        type=float,
        required=False
    )
    parser.add_argument(
        '-vb',
        '--verbose',
        help='Verbose output.',
        required=False,
        choices=['True', 'False']
        
    )
    parser.add_argument(
        '-ran',
        '--random',        
        help='Randomize hyper parameters.',
        required=False,
        default='False',
        choices=['True', 'False']
        
    )
    parser.add_argument(
        '-loop',
        '--looping',
        help='Enables looping of sklearn models (useful for random hyper parameter)',
        default='False',
        required=False,
        choices=['True', 'False']
    )
    parser.add_argument(
        '-test',
        '--eval-test',
        help='Evaluates the model on the test dataset.',
        default='False',
        required=False,
        choices=['True', 'False']
    )
    parser.add_argument(
        '-output-file',
        help="Name of output file. Default will gen a new name.",
        required=False,
        default=None)
    


def main():
    """
    Collects arguments and configurations, or sets defaults.
    Calls either the NER, analysis, formatting, or testing routines.

    Args:
        None

    Returns:
        None

    Raises:
        None

    """

    parser = argparse.ArgumentParser()
    add_args(parser)
    args = parser.parse_args()

    config_path = args.config or \
                  os.environ.get('SNER_CONF') or \
                  'sner.conf'

    if os.path.exists(config_path):
        config_file = json.load(open(config_path))
    else:
        config_file = {
            'run': '',
            'path': '',
            'corpus': '',
            'attestations': '',
            'seed-rules': '',
            'iterations': '',
            'max-rules': '',
            'alpha': '',
            'k': '',
            'norm': {
                'date': '',
                'geo': '',
                'num': '',
                'prof': ''
            },
            'params': {
                'alpha': '',
                'C': '',
                'criterion': '',
                'degree': '',
                'kernel': '',
                'loss': '',
                'max_features': '',
                'max_depth': '',
                'min_samples_split': '',
                'max_leaf_nodes': '',
                'min_samples_leaf': '',
                'n_estimators': '',
                'penalty': '',
                'splitter': ''
            }
        }

    config = {
        'run': args.run or \
               config_file['run'] or \
               'ner',
        'path': args.path or \
                config_file['path'] or \
                'data/',
        'corpus': args.corpus or \
                  config_file['corpus'] or \
                  'corpus.csv',
        'output-file': args.output_file,
        'attestations': args.attestations or \
                        config_file['attestations'] or \
                        'attestations.csv',
        'seed-rules': args.seed_rules or \
                      config_file['seed-rules'] or \
                      'seed_rules.csv',
        'iterations': args.iterations or \
                      config_file['iterations'] or \
                      5,
        'max-rules': args.max_rules or \
                     config_file['max-rules'] or \
                     5,
        'alpha': args.alpha or \
                 config_file['alpha'] or \
                 0.1,
        'k': args.k or \
             config_file['k'] or \
             2.0,        
        'train_size': args.train_size or \
             0.7,
        'dev_size': args.dev_size or \
             0.3,
        'looping': bool_check(
            args.looping,
            None,
            False
        ),
        'eval-test': bool_check(
            args.eval_test,
            None,
            False
        ),
        'flags': {
            'disjoint_names': bool_check(
                args.disjoint_names,
                None,
                False
            ),
            'use_left_context': bool_check(
                 args.export_left_rules,
                None,
                False
            ),            
            'use_right_context': bool_check(
                args.export_right_rules,
                None,
                False
            ),
            'use_spelling': bool_check(
                args.export_spelling_rules,
                None,
                False
            ),
            'use_syllables': bool_check(
                 args.export_found,
                None,
                True
            ),
            'verbose': bool_check(
                 args.verbose,
                None,
                False
            ),
            'use_atf': bool_check(
                 args.use_atf,
                None,
                False
            )
        },
        'norm': {
            'date': bool_check(
                args.norm_date,
                config_file['norm']['date'],
                False
            ),
            'geo': bool_check(
                args.norm_geo,
                config_file['norm']['geo'],
                False
            ),
            'num': bool_check(
                args.norm_num,
                config_file['norm']['num'],
                True
            ),
            'prof': bool_check(
                args.norm_prof,
                config_file['norm']['prof'],
                True
            )
        },        
        'params': {
            'alpha': config_file['params']['alpha'] or \
                     0.503706954708716,
            'C': config_file['params']['C'] or \
                 1.67640892878145,
            'criterion': config_file['params']['criterion'] or \
                         'entropy',
            'degree': config_file['params']['degree'] or \
                      4,
            'kernel': config_file['params']['kernel'] or \
                      'linear',
            'loss': config_file['params']['loss'] or \
                    'squared_hinge',
            'max_features': config_file['params']['max_features'] or \
                            75,
            'max_depth': config_file['params']['max_depth'] or \
                         165,
            'min_samples_split': config_file['params']['min_samples_split'] or \
                                 2,
            'max_leaf_nodes': config_file['params']['max_leaf_nodes'] or \
                              None,
            'min_samples_leaf': config_file['params']['min_samples_leaf'] or \
                                1,
            'n_estimators': config_file['params']['n_estimators'] or \
                            100,
            'penalty': config_file['params']['penalty'] or \
                       'elasticnet',
            'splitter': config_file['params']['splitter'] or \
                        'best'
        }
    }

    config['params']['random'] = False
    if args.random == 'True':
        config['params']['random'] = True

    # Routines
    if config['run'] == 'analysis':
        # analysis.main(config)
        pass
    elif config['run'] == 'export':
        export.main(config)
    elif config['run'] == 'export-atf':
        export_atf.main(config)
    elif config['run'] == 'formatting':
        #formatting.main(config)
        pass
    elif config['run'] == 'testing':
        #pytest.main(['tests/'])
        print("no.")
    elif config['run'] == 'over-fit':
        overfit_check.main(config)
    elif config['run'] == 'output':
        output.main(config)

    # Models
    else:
        sklearn_launcher.main(config)


if __name__ == '__main__':
    main()



import random
import codecs
import operator

unique_names = False


def main(config):
    global unique_names    
    path = config['path']
    corpus = config['corpus']

    use_atf = config['flags']['use_atf']
    
    train_key = path + 'target_train.KEY'
    train_target = path + 'target_train.RT'
    dev_key = path + 'target_dev.KEY'
    dev_target = path + 'target_dev.RT'
    dev_pred = path + 'dev_prediction.RT'

    atf_key = path + 'target_atf.KEY'
    atf_pred = path + 'atf_prediction.RT'

    flags = config['flags']
    if flags['disjoint_names'] == True:
        uniqueNames = True
        print("Using disjoint name sets for train / dev.")
    else:
        uniqueNames = False

    

    train_names = set()
    dev_names = set()
    pred_names = set()
    dev_output = []
    
    atf_names = set()
    atf_output = []
    
    
    total_train_names = addNames(train_key, train_target, train_names)
    total_dev_names = add_names_and_unique(dev_key, dev_target, dev_names, train_names, dev_output)
    total_pred_names = addNames(dev_key, dev_pred, pred_names)

    total_atf_names = None
    if use_atf:
        total_atf_names = add_names_and_unique(atf_key, atf_pred,
                                               atf_names, train_names, atf_output, 2)

    dev_unique_names = dev_names - train_names
    new_names = pred_names - train_names    
    atf_unique_names = None
    if use_atf:
         atf_unique_names = atf_names - train_names

    print ("Training found %d names." % len(train_names))

    if not use_atf:
        print("Test found %d names." % len(dev_names))
        print("Unique names in test: %d" % (len(new_names)))
        print("Test name occurrences: %d" % total_dev_names)
        print("Test unique name occurrences: %d - %.3f%%" % (len(dev_output), 100 * len(dev_output) / total_dev_names))
        
    # Exit early when looking at ATF names.             
    if use_atf:
        print("Names found in atf: %d" % len(atf_names))
        print("Unique names found in atf: %d" % len(atf_unique_names))
        print("Total name occurrences in ATF: %d" % total_atf_names)
        print("Unique name occurrences in ATF: %d - %.3f%%" % (len(atf_output), 100 * len(atf_output) / total_atf_names ))
        outputATF(config, atf_output)
        return


        
    dev_key = open(dev_key, "r")
    dev_target = open(dev_target, "r")
    dev_pred = open(dev_pred, "r")

    correct = 0
    missed = 0
    mislabeled = 0
    correct_names = set()

    verbose = config['flags']['verbose']

    for x, y, z in zip(dev_key, dev_target, dev_pred):
        x = x.split(',')[3].strip()
        y = y.strip()
        z = z.strip()
        
        if int(y) == 1:
            if int(z) == 1:
                if x in new_names:
                    if verbose:
                        print("Found correct new name: %s" % x)
                    correct_names.add(x)
                    correct += 1
            else :
                if verbose:
                    print("Missed the correct name: %s" % x)
                missed += 1

        else:
            if int(z) == 1:
                if verbose:
                    print("Mislabeled name: %s" % x)
                mislabeled += 1


    
    print ()
    print ("Total names in train: %d\nTotal names in test: %d\nTotal names in pred: %d\n" % (total_train_names, total_dev_names, total_pred_names))
    
    print ("Training has %d unique names.\nTest has %d unique names.\nUnique names to test: %d\nUnique names found in pred: %d"
           % (len(train_names), len(dev_names), len(dev_unique_names), len(new_names)))
    print ("Correct unique names in pred: %d / %d [%.2f%%]\n" % (len(correct_names), len(dev_unique_names), 100 * len(correct_names) / len(dev_unique_names)))
    
    print ("Correct unique occurrences: %d" % (correct))
    print ("Missed names: %d / %d [%.2f%%]" % (missed, total_dev_names, 100 * missed / total_dev_names))
    print ("Mislabeled names: %d / %d [%.2f%%]" % (mislabeled, total_pred_names, 100 * mislabeled / total_pred_names))


def outputATF(config, output_list):
    path = config['path']
    count = 0
    unique_names = set()
    selected = []

    name_counts = {}
    name_line = {}

    for x in output_list:
        name = x.split(',')[2].strip()
        if name not in name_counts.keys():
            name_counts[name] = 0
        name_counts[name] = name_counts[name] + 1
        name_line[name] = x

    sorted_names = sorted(name_counts.items(), key=operator.itemgetter(1), reverse=True)

    end = len(sorted_names)
    while count < end:
        #print(sorted_names[count])        
        line = name_line[sorted_names[count][0]]
        #print(line)
        #print()
        result = line.split(',')
        if len(result) > 3:
            combo = ','.join(result[2:])
            result[2:] = []
            result.append(combo)
        result.append(sorted_names[count][1])
        selected.append(result)
        count += 1
    
    #while count < 3:
    #    line = random.choice(output_list)
    #    name = line.split(',')[2].strip()
    #    if name not in unique_names:
    #        unique_names.add(name)
    #        selected.append(line.split(','))
    #        count += 1

    print("----------------------")
    print("lineid, wordid, occurences, name")
    ordered = sorted(selected, key=lambda x: float(x[0]))

    corpus = codecs.open(path + config['corpus'], 'r', encoding = 'utf-8')

    lineid = -1
    nextname = 0
    len_names = len(ordered)
    for line in corpus:
              
        lineid += 1
        
        while  nextname < len_names and lineid == int(ordered[nextname][0]):
            nn = ordered[nextname]
            print("%s, %s, %s, %s" % (nn[0].strip(), nn[1].strip(), nn[3], nn[2].strip()))
            if config['flags']['verbose']:
                print(line.strip())
                print()
            nextname += 1
        
            

        if nextname >= len(ordered):
            break


    print(lineid)
    print(nextname)
    print(len(ordered))


def numeric_compare(x, y):
    return int(x) - int(y)

def add_names_and_unique(file_key, file_target, names, train_names, output_list, index=3):
    global unique_names
    key = open(file_key, "r")
    target = open(file_target, "r")
    name_count = 0
    for x, y in zip(key, target):
        x = x.strip()
        y = y.strip()
        if int(y) == 1:
            name_count += 1
            name = x.split(',')[index].strip()
            names.add(name)
            if name not in train_names:              
                output_list.append(x)

    return name_count

    

def addNames(file_key, file_target, names):
    key = open(file_key, "r")
    target = open(file_target, "r")
    name_count = 0
    for x, y in zip(key, target):
        x = x.strip()
        y = y.strip()
        if int(y) == 1:
            name_count += 1
            name = x.split(',')[3].strip()
            names.add(name)

    return name_count






from sner.classes import Rule, Token
from sner.scripts.ner import rulesperformance, rulefilter


def main(corpus, rules, names, max_rules, iteration, options, display):
    """
    This is meant to generate conextual rules from a set of identified
    name tokens. It needs the corpus as well as the name set in order to
    assess the performance of any rules it finds from the names

    Args:
        corpus (set): Set of all Token objects in the corpus.
        rules (set): Set of all Rule objects that have been found so far.
        names (set): Set of Token objects to derive new context rules from.
        max_rules (int): maximum number of rules to be accepted each iteration.
        iteration (int): what iteration is the algorithm currently on?
        options (Options): collection of configuration options

    Returns:
        new_rules (set): Set of Rule objects.

    Raises:
        None

    """

    new_rules = set()

    for name in names:
        # New rules are assumed to have no strength until they are assessed
        #  later by scanning the whole corpus

        # Create left context rule
        left_context = Rule(Rule.Type.left_context, str(name.left_context), -1)
        # Create right context rule
        right_context = Rule(Rule.Type.right_context, str(name.right_context), -1)

        # No redundant rules allowed!
        if left_context not in rules:
            left_context.iteration = iteration
            new_rules.add(left_context)

        if right_context not in rules:
            right_context.iteration = iteration
            new_rules.add(right_context)

    rulesperformance.main(corpus, new_rules, options, iteration, display)

    new_rules = rulefilter.main(new_rules, max_rules)

    return new_rules



from sner.classes import Rule, Token
from sner.scripts.ner import namesfromrule


def main(tokens, rules):
    """
    This will update the strength of all tokens it is given, using
        the rules it is given.
    Args:
        tokens (set): Set object of Token objects.
        rules (set): Set object of Rule objects.

    Returns:
        None

    Raises:
        ValueError

    """

    for rule in rules:
        names = namesfromrule.main(tokens, rule)

        for name in names:
            # Raise ValueError("attempted to apply same rule to a token twice!")

            # Use statistical rules to calculate the new probability
            # In other words, the probability that all other applicable rules in addition
            #  to the current one are wrong
            initialprob = name.name_probability

            if ((initialprob < 0) and (initialprob > 1)):
                raise ValueError("Token \"" + str(name) + "\" has impossible name probability (v < 0 or v > 1): " + str(initialprob))

            addedprob = rule.strength

            if ((addedprob < 0) and (addedprob > 1)):
                raise ValueError("Rule \"" + str(rule) + "\" has impossible strength rating (str < 0 or str > 1): " + str(addedprob))

            newprob = 1 - ((1 - initialprob) * (1 - addedprob))

            if ((newprob < 0) and (newprob > 1)):
                raise ValueError("Generated impossible name probability: " + str(newprob))

            name.name_probability = newprob



# This file is meant to generate spelling rules from a list of identified name
# tokens.

from sner.classes import Rule, Token
from sner.scripts.ner import rulefilter, rulesperformance


# Produces a list of monograms on up to k-grams
def getKgrams(names, k):
    """
    Produces lists of grams, from monograms up to k-grams.
    Duplicate found in /scripts/readnames.py
    Example use to get monograms, birgams, and trigrams:
    getKgrams(getPNs(), 3)
    Args:
        names (dict) = dictionary of the form { Names : Occurrences }
        k (int) = Max k-gram you want to retrieve.
            Special values of k = -1 returning dictionary of monograms.
                          and k = -2 returning dictionary of monograms,
                                  followed by a dictionary of bigrams.

    Returns:
        A dictionary of all grams up to order k

    Raises:
        None

    """

    grams = list()

    for name in names:
        syllables = str(name).split('-')

        for i in range(1, k + 1):
            end = len(syllables) - i + 1
            for j in range(0, end):
                gram = '-'.join(syllables[j: j + i])
                grams.append(gram)

    return grams


def gramsToRules(kgrams, allrules, iteration):
    rules = set()

    for gram in kgrams:
        rule = Rule(Rule.Type.spelling, gram, -1)
        if not rule in allrules:
            rules.add(rule)
            rule.iteration = iteration

    return rules


def main(corpus, allrules, names, maxrules, iteration, options, display):
    """
    Produces lists of grams, from monograms up to k-grams.
    Duplicate found in /scripts/readnames.py
    Example use to get monograms, birgams, and trigrams:
    getKgrams(getPNs(), 3)
    Args:
        corpus (set) = set of all tokens in the corpus
        allrules (set) = set of all rules in the corpus
		names (set) = set of tokens that will be used to generate new spelling rules
        maxrules (int) = maximum number of rules to be accepted with each iteration
        iteration (int) = what iteration are we currently on?
        options (Options) = collection of configuration data

    Returns:
        Set of spelling rules generated from names

    Raises:
        None

    """
    # Uses up to trigrams currently
    maxGram = 3

    # Get all possible spelling rules from the names passed in
    kgrams = getKgrams(names, maxGram)
    rules = gramsToRules(kgrams, allrules, iteration)

    rulesperformance.main(corpus, rules, options, iteration, display)

    rules = rulefilter.main(rules, maxrules)

    return rules



from sner.scripts import professions
import codecs
import re


def get_counts(data):
    """
    This function gets the total occurrences of words and syllables in the
    original Unicode Garshana corpus.  To do this, it opens a .csv file with
    utf-16 encoding, and splits on commans, expecting the line of sumerian text
    to be in the 8th column. Filters annotations from each line, and tracks the
    occurrence of each word and syllable. All combinations of unigrams, bigrams,
    and trigrams are treated as individual syllables.

    Args:
        data = filename of the corpus .csv file, consistent with the formatting
			   of the .csv files provided with the Garshana corpus.

    Returns:
        Returns dictionary of the number of times any unique word occurs, as
        well as a dictionary of occurrences for syllables.

    Raises:
        IOError

    """

    word_count = {}
    syll_count = {}

    infile = data.corpus
    try:

        open_file = codecs.open(infile, 'r', encoding='utf-16')
        for line in open_file:
            line = line.lower()
            # Remove tablet indexing info and line numbers. Grab only text data
            line = line.split(',')
            text = clean_line(line[7])

            # Update the occurrences of the words in the line
            for word in text.split():
                count = word_count.setdefault(word, 0)
                word_count[word] = count + 1

                # Track occurrences of syllables
                update_syllable_count(word, syll_count)

        open_file.close()
    except IOError:
        print("Cannot open: " + infile)

    return (word_count, syll_count)


def update_syllable_count(word, syll_count):
    """
    Update the total occurrence counts of each unigram, bigram, and trigram
    syllable that occurs in the word.  Note: syllables are separated by a
    dash ('-').
    """

    syllables = word.split('-')
    for i in range(1, 4):
        for j in range(len(syllables) - i + 1):
            gram = '-'.join(syllables[j: j + i])
            count = syll_count.setdefault(gram, 0)
            syll_count[gram] = count + 1


def clean_line(line, normNum=True, normProf=True):
    """
    Clean a line of data, removing all annotations from the line.

    NOTE: The line is expected to only be the TEXT portion of the data files.
    I.e. the ID and line number parts of the data files are expected to be
    previously removed.

    Args:
        line (str): Line of just the text section of a tablet
    Returns:
        line (str): The line, with all annotations removed

    Raises:
        None
    """

    # Remove square brackets, ceiling characters, question marks, other
    # questionable characters, and line breaks
    line = re.sub(r'(\[|\])', '', line)
    line = re.sub(r'(⌈|⌉)', '', line)
    line = re.sub(r'( / )', ' ', line)
    line = re.sub(r'/', '', line)
    line = re.sub(r'\?', '', line)
    line = re.sub(r'([<]|[>])+', '', line)
    line = re.sub(r'!', '', line)
    line = re.sub(r'"', '', line)

    # Remove researcher's notes, and multiple dashes or '='s
    line = re.sub(r'(\(.*\))', '', line)
    line = re.sub(r'(#[.]*)', '', line)
    line = re.sub(r'[-]{2}', '', line)
    line = re.sub(r'[=]{2}', '', line)

    # Replace numbers with 'number'
    if normNum is True:
        line = re.sub(r'\b(?<!-)(\d+)(?![\w-])', 'number', line)
        line = re.sub(r'[-+]?\b\d+\b', 'number', line)

        #line = re.sub(r'\b([\-\.0-9]+)(?![\w-])', 'number', line)

    # Replace professions with 'profession'
    if normProf is True:
        line = professions.replaceProfessions(line)

    # Remove blank character at end of line
    linelength = len(line)
    if (linelength > 0 and line[linelength-1] == ""):
        del line[0:linelength-2]

    return line


def main():
    # Testing
    word_count, syll_count = get_counts()
    print(word_count)
    print()
    print(syll_count)

if __name__ == '__main__':
    main()



"""Imports the corpus data for the scikitlearn model"""
import math
import pandas as pd
from sner.classes import Display


def main(config):
    """
	Loads the corpus data from a file and returns it as a dataframe.  The
	DataFrame stores lines of corpus data as well as various metadata
	specified by the cols variable defined below.  Able to handle .csv formats
	as well as .atf formats.

    Args:
        config(dict) = contains file locations and runtime flags, such as the
					   location of the corpus file.

    Returns:
        corpus(pandas.DataFrame) = corpus information represented in a
								   dataframe format.

    Raises:
        None

    """

    file_name = config['corpus']
    path = config['path']

    if file_name[-3:] == 'atf':
        sep = 'READ_AS_ONE_COLUMN'
        header = None
        engine = 'python'
    elif file_name[-3:] == 'csv':
        sep = ','
        header = 0
        engine = 'c'

    reader = pd.read_table(
        filepath_or_buffer=path + file_name,
        sep=sep,
        header=header,
        engine=engine
    )

    cols = {
        'BDTNS ID'   : [],
        'Line ID'    : [],
        'CDLI ID'    : [],
        'Designation': [],
        'Q ID'       : [],
        'Q Line'     : [],
        'Language'   : [],
        'Object'     : [],
        'Surface'    : [],
        'Discourse'  : [],
        'Line'       : [],
        'Text'       : [],
        'Translation': []
    }

    corpus = pd.DataFrame(data=cols)

    objects = [
        'tablet',
        'envelope',
        'prism',
        'bulla',
        'fragment'
    ]

    discourse = [
        'catchline',
        'colophon',
        'date'
        'signatures',
        'signature',
        'summary',
        'witness'
    ]

    def read_csv(corpus, reader):
		"""
		Handles loading .csv format for the container function.

		Args:
			corpus(pandas.DataFrame) = dataframe representation of the corpus
									   to be filled out
			reader(pandas.DataFrame) = contains raw file data scanned into
									   a dataframe

		Returns:
			corpus(pandas.DataFrame) = corpus information represented in a
									   dataframe format.

		Raises:
			None

		"""

        bdtns_id = ''
        line_id = ''
        cdli_id = ''
        name = ''
        q_id = ''
        q_line = ''
        lang = ''
        obj = ''
        surf = ''
        disc = ''
        line = ''
        text = ''
        tran = ''

        i = 0
        for index, row in reader.iterrows():
            if row['Text'] != 'NaN':
                i += 1
                bdtns_id = row['Id BDTNS']
                line_id = row['Id Line']
                line = row['Line']
                text = row['Text']
                corpus.loc[i, 'BDTNS ID'] = bdtns_id
                corpus.loc[i, 'Line ID'] = line_id
                corpus.loc[i, 'CDLI ID'] = cdli_id
                corpus.loc[i, 'Designation'] = name
                corpus.loc[i, 'Q ID'] = q_id
                corpus.loc[i, 'Q Line'] = q_line
                corpus.loc[i, 'Language'] = lang
                corpus.loc[i, 'Object'] = obj
                corpus.loc[i, 'Surface'] = surf
                corpus.loc[i, 'Discourse'] = disc
                corpus.loc[i, 'Line'] = line
                corpus.loc[i, 'Text'] = text
                corpus.loc[i, 'Translation'] = tran

        return corpus


    def read_atf(corpus, reader):
		"""
		Handles loading .atf format for the container function.

		Args:
			corpus(pandas.DataFrame) = dataframe representation of the corpus
									   to be filled out
			reader(pandas.DataFrame) = contains raw file data scanned into
									   a dataframe

		Returns:
			corpus(pandas.DataFrame) = corpus information represented in a
									   dataframe format.

		Raises:
			None

		"""

        bdtns_id = ''
        line_id = ''
        cdli_id = ''
        name = ''
        q_id = ''
        q_line = ''
        lang = ''
        obj = ''
        surf = ''
        disc = ''
        line = ''
        text = ''
        tran = ''

        display = Display()
        display.start('Importing corpus')

        i = 0
        for index, row in reader.iterrows():
            display.update_progress_bar(index + 1, len(reader))
            if row[0][0] == '&':
                cdli_id = row[0][:8]
                name = row[0][11:]
                q_id = ''
                q_line = ''
                lang = ''
                obj = ''
                surf = ''
                disc = ''
                line = ''
                text = ''
                tran = ''
            elif row[0][0] == '#':
                if row[0][1:10] == 'atf: lang':
                    lang = row[0][11:]
                if row[0][1:4] == 'tr.':
                    tran = row[0][4:]
                else:
                    # Ignore Comments
                    pass
            elif row[0][0] == '@':
                if row[0][1:] in objects:
                    obj = row[0][1:]
                elif row[0][1:7] == 'object':
                    obj = row[0][8:]
                elif row[0][1:] in discourse:
                    disc = row[0][1:]
                else:
                    surf = row[0][1:]
            elif row[0][0] == '$':
                # Ignore States
                pass
            elif row[0][:3] == '>>Q':
                q_id = row[0][2:9]
                q_line = row[0][10:]
            else:
                if row[0][0].isdigit():
                    i += 1
                    period = row[0].index('.')
                    line = row[0][:period]
                    text = row[0][period + 2:]
                else:
                    text += row[0]

                corpus.loc[i, 'BDTNS ID'] = bdtns_id
                corpus.loc[i, 'Line ID'] = line_id
                corpus.loc[i, 'CDLI ID'] = cdli_id
                corpus.loc[i, 'Designation'] = name
                corpus.loc[i, 'Q ID'] = q_id
                corpus.loc[i, 'Q Line'] = q_line
                corpus.loc[i, 'Language'] = lang
                corpus.loc[i, 'Object'] = obj
                corpus.loc[i, 'Surface'] = surf
                corpus.loc[i, 'Discourse'] = disc
                corpus.loc[i, 'Line'] = line
                corpus.loc[i, 'Text'] = text
                corpus.loc[i, 'Translation'] = tran

        display.finish()
        return corpus


    if file_name[-3:] == 'atf':
        corpus = read_atf(corpus, reader)
    elif file_name[-3:] == 'csv':
        corpus = read_csv(corpus, reader)

    return corpus



from sner.scripts import professions, utilities
import codecs
import csv
import re
import argparse
import random


uniqueNames = True

# To run:
# python3 sner.py -r export -nn True -np True


def findKnown(config, known_pn, known_gn):
    """
    Iterates through the attenstations file for personal and geographical names
     and adds the lineIDs in a list to the known_pn and known_gn dictionaries.
    Args:
        data (?)
        options (?)
        known_pn = { Personal name : line ID }, to be filled in.
        known_gn = { Geographical name : line ID}, to be filled in.

    Returns:
        Updates known_pn and known_gn dictionaries with the lineIDs.

    Raises:
        None

    """
    path = config['path']
    norm_num = config['norm']['num']
    norm_prof = config['norm']['prof']

    file = codecs.open(path + 'attestations.csv', 'r', encoding = 'utf-8')
    # find all of the names
    for line in file:
        line = line.split(',')
        text = utilities.clean_line(line[4].rstrip(), norm_num, norm_prof)
        text = text.lower()
        name = utilities.clean_line(line[5].rstrip(), norm_num, norm_prof)
        name = name.lower()
        line_id = line[1]
        line_id = re.sub("[L]", "", line_id)

        if line[9].rstrip() == 'PN':
            if (name not in known_pn):
                known_pn[name] = [line_id]
            else :
                known_pn[name].append(line_id)
        if line[9].rstrip() == 'GN':
            if (name not in known_gn):
                known_gn[name] = [line_id]
                if (name in known_pn):
                    print ("Note: ", name, " is both a PN and GN.")
            else :
                known_gn[name].append(line_id)


left_features = []
left_features_phase1 =  ['dumu','giri3', 'iti',
                  'u3',
                  'arad2',
                  'gurum2',
                  'en-nu',
                  'zi-ga',
                  'igi',
                  'im-sumur',
                  'in-u',
                  'du-u2-um',
                  'i7',
                  'gu2-gal',
                  'a2-bi',
                  'a-ga2-ri-na',
                  'puzur4-a-ku-um',
                  'tu7',
                  'nig2-gu2-na',
                  '{giš}garigx',
                  'nig2-lam2',
                  'sag',
                  'ki-a-nag',
                  '{tug2}nig2-lam2-ab',
                  '{na4}profession',
                  '{gi}dur-gazi',
                  'nig2-kas7',
                  'bara2',
                  'i3-giš-bara2',
                  'kaskal',
                  'gar3-da-hi-še3',
                  'kab2',
                  'tug2-ba',
                  '{giš}u3-suh5-sig',
                  '{gi}ma-an-sim',
                  'nig2-ar3-ra',
                  '{gi}gur5',
                  'udug',
                  '{tug2}bar-dul5-ab']
right_features = []
right_features_phase1 = ['lugal-e','nubanda3',
                  'dub-sar',
                  'arad2-zu',
                  'ašgab',
                  'nu-banda3',
                  'a-zu',
                  'e2-šim×gar',
                  'al-tar',
                  'a2-ag2-ga2',
                  'geme2-arad2',
                  'al-nag4-ga2',
                  '{d}nin-{d}si4-an-na',
                  '1-kam',
                  '2-kam',
                  'a-ra2',
                  'e2-uš-bar',
                  'lu2-kin-gi4-a',
                  'e2-da-na{ki}-še3',
                  '3-kam',
                  'ba-ra-šeg6',
                  '{dug}utul2',
                  'e2-kišib-ba-še3',
                  'ša3-du',
                  'hun-ga2-e-ne',
                  'ka',
                  'ga2-nun-še3',
                  'ki-a-nag',
                  'ki-mu-ra',
                  'si-ga',
                  'e2-uš-bar-še3',
                  '4-kam',
                  'diri-ga',
                  'kurušda',
                  'alan',
                  '1-sila3',
                  'ma-ma-šar-ra-at',
                  'za3-gu2-la2',
                  'du11-ga',
                  '{d}šu-{d}suen-e-te-el2-ka-{d}en-lil2',
                  'lu2-kas4',
                  'šuš3']

spelling_features = []
spelling_features_phase1 = ['i3-li',
                     'i3-li2',
                     'kab-ta',
                     'kab-ta2',
                     'a-ra',
                     'ra-a',
                     'a-bi',
                     '{d}nin-kar',
                     'ra-am',
                     'kar3',
                     'ta2-kar3',
                     'am-i3',
                     'ra-am-i3',
                     'na-ah',
                     'sig5',
                     'ah-i3-li2',
                     'na-ah-i3',
                     'ah-i3',
                     'na-ra',
                     'na-ra-am',
                     'li2-a',
                     'bi2-a',
                     'a-li2',
                     'li2-i3',
                     'a-me',
                     'ar3-ra',
                     'nig2-ar3',
                     'nig2-ar3-ra',
                     'ra-sig5',
                     'ar3-ra-sig5',
                     'ba-ba',
                     'li2-bi',
                     'la-ni',
                     'sa2-a',
                     'bi-la',
                     'ra-ni',
                     'bi-la-ni',
                     'li2-bi-la',
                     'at-i3',
                     'ki-sa2',
                     'ta2-ki',
                     'si-ru',
                     'si-ru-um',
                     'li2-um',
                     'li2-an.dul3',
                     'duru5-sig5',
                     'li2-aš-ra',
                     'aš-ra-ni',
                     'aš-ra',
                     'li2-aš',
                     'ki-ru',
                     'ta2-ki-ru',
                     'ki-ru-um',
                     'ba-ra',
                     'gašam-e',
                     'gašam-e-ne',
                     'ba-ak',
                     'duru5-kab-bi',
                     'kab-bi',
                     'duru5-kab',
                     'li2-um{ki}',
                     'ur2-i3',
                     'uz-ra-ni',
                     'za-gar3',
                     'sa2-ab',
                     'li2-uz-ra',
                     'li2-uz',
                     'uz-ra',
                     'ši-ru',
                     'bi2-za',
                     'gur11',
                     'u4-11',
                     'li2-tu',
                     'ga-li2']

known_symbols = []

test_names = set()
dev_names = set()

def writeSparse(config, out_features, word_left, word_middle, word_right, x_index):
    """
    Writes a single x vector of features in a one hot inspired representation
      to the out_features file.

    Args:
      out_features = output file to write features
      word_left = the word left of the word to be output
      word_middle = the word in question
      word_right = the right context of the word to be output
      x_index = the row id for the feature entry

    Returns:
      Nothing

    Raises:
      None

    """

    offset = 0
    for i in range(len(left_features)):
        left = left_features[i]
        if word_left == left:
            out_features.write("{} {} 1\n".format(x_index, i + offset))

    offset = len(left_features)
    for i in range(len(right_features)):
        right = right_features[i]

        if word_right == right:
            out_features.write("{} {} 1\n".format(x_index, i + offset))

    offset = len(left_features) + len(right_features)
    for i in range(len(spelling_features)):
        spelling = spelling_features[i]
        if spelling in word_middle:
            out_features.write("{} {} 1\n".format(x_index, i + offset))

    # write number of syllables
    offset = len(left_features) + len(right_features) + len(spelling_features)
    symbols = word_middle.split('-')
    out_features.write("{} {} {}\n".format(x_index, offset, len(symbols)))
    offset += 1

    if config['flags']['use_syllables'] == False:
        return

    foundSym = []
    for sym in symbols:
        if sym not in known_symbols:
            known_symbols.append(sym)
        if sym not in foundSym:
            foundSym.append(sym)
            out_features.write("{} {} 1\n".format(x_index, known_symbols.index(sym) + offset))



def main(config):
    """
    Finds all names in options.attestations file, and goes through each word
      in the main options.corpus file to create a sparse matrix file to be used
      with scikit learn.
    Args:
        options.norm_date = True to normalize numbers
        options.norm_num = True to normalize numbers
        options.norm_prof = True to normalize professions
        options.norm_geo = True to normalize geographical names

    Returns:
        .csv file with the format:
            TabletID, LineID, LocationInSentence, Word, Word Type
        Word Types: '-' for unknowns.
                    'PN' to identify personal names.
                    'GN' to identify geographical names.
                    'PF' to identify professions.

    Raises:
        None

    """
    global left_features
    global right_features
    global spelling_features
    global uniqueNames
    path = config['path']
    corpus = config['corpus']

    if config['norm']['date']:
        print("Normalizing dates!")
    if config['norm']['geo']:
        print("Normalizing geo names!")
    if config['norm']['num']:
        print("Normalizing numbers.")
    if config['norm']['prof']:
        print("Normalizing professions.")

    flags = config['flags']
    if flags['disjoint_names'] == True:
        uniqueNames = True
        print("Using disjoint name sets for train / dev.")
    else:
        uniqueNames = False

    if flags['use_left_context'] == True:
        left_features = left_features_phase1
        print("Using left context from phase 1.")
    else:
        left_features = []

    if flags['use_right_context'] == True:
        print("Using right context from phase 1.")
        right_features = right_features_phase1
    else:
        right_features = []

    if flags['use_spelling'] == True:
        print("Using spelling rules from phase 1.")
        spelling_features = spelling_features_phase1
    else:
        spelling_features = []

    if flags['use_syllables'] == True:
        print("Using unique syllables as features.")


    known_pn = {}
    known_gn = {}
    findKnown(config, known_pn, known_gn)

    # iterate each line in Garshana Text and output to specified file
    file2 = codecs.open(path + corpus, 'r', encoding = 'utf-8')


    dev_size = config['dev_size']
    train_size = config['train_size']

    if dev_size + train_size > 1.001 or dev_size < 0 or train_size < 0:
        print(dev_size, train_size)
        print("Invalid dev or train size! Expecting both to be 0-1 and dev+train <= 1.")
        return


    test_size = 1 - (dev_size + train_size)

    lines = file2.read().splitlines()
    end_train =  len(lines) * train_size
    end_dev = len(lines) * dev_size
    end_test = len(lines) * test_size


    print ("Train size: %.2f%% %d lines" % (train_size,end_train))
    print ("Dev size: %.2f%% %d lines" % (dev_size,end_dev))
    print ("Test size: %.2f%% %d lines" % (test_size,end_test))

    out_features = open('data/features_train.sparseX', "w")
    out_target = open('data/target_train.RT', "w")
    out_key = open('data/target_train.KEY', "w")

    out_features2 = open('data/features_dev.sparseX', "w")
    out_target2 = open('data/target_dev.RT', "w")
    out_key2 = open('data/target_dev.KEY', "w")

    out_features3 = open('data/features_test.sparseX', "w")
    out_target3 = open('data/target_test.RT', "w")
    out_key3 = open('data/target_test.KEY', "w")


    x_index = 0
    x_index2 = 0
    x_index3 = 0

    train_lines = 0
    dev_lines = 0
    test_lines = 0


    while len(lines) > 0:
        if (train_lines < end_train or (dev_lines >= end_dev and test_lines >= end_test)):
            line = random.choice(lines)
            x_index = writeLine(x_index, config, line, out_features, out_target, out_key, known_pn, known_gn, True)
            lines.remove(line)
            train_lines += 1
        if (dev_lines < end_dev and len(lines) > 0):
            line = random.choice(lines)
            x_index2 = writeLine(x_index2, config, line, out_features2, out_target2, out_key2, known_pn, known_gn, False)
            lines.remove(line)
            dev_lines += 1
        if (test_lines < end_test and len(lines) > 0):
            line = random.choice(lines)
            x_index3 = writeLine(x_index3, config, line, out_features3, out_target3, out_key3, known_pn, known_gn, False)
            lines.remove(line)
            test_lines += 1


    writeKey(path)


def writeLine(x_index, config, line, out_features, out_target, out_key, known_pn, known_gn, test_run):
    norm_date = config['norm']['date']
    norm_geo  = config['norm']['geo']
    norm_num  = config['norm']['num']
    norm_prof = config['norm']['prof']


    line = line.split(',')
    if len(line) < 8:
        return x_index
    tablet_id = line[0]

    # Clean up the line and ensure that there are no extra spaces
    text = utilities.clean_line(line[7].rstrip(), norm_num, norm_prof)
    text = text.lower()
    text = " ".join(text.split())
    text = text.strip()

    # We should skip the first line that labels the csv
    if "text" in text:
        return x_index

    newText = text;
    words = text.split(' ')
    line_id = line[6]
    # skip description lines
    if line_id == "":
        return x_index
    # change to line_id
    line_id = line[4]
    line_id = re.sub("[L]", "", line_id)
    last_word = ""
    last_word_2 = ""
    last_line_id = ""
    last_tablet = ""
    last_index = -1

    last_name = False
    last_geo = False

    if "'" in line_id:
        line_id = re.sub("[']", "", line_id)
        line_id = "'" + line_id

    if ("{joined" in text):
        print(text)
        return x_index

    if ("writen" in text):
        print(text)
        return x_index

    if ("erasure" in text):
        print("Getting rid of erasure: %s" % text)
        return x_index



    for i in range(len(words)):
        w = words[i]
        w_type = "-"

        if ("." in w):
            continue

        # Set types if they are known and avoid
        # conflicts by using the line_id.
        if w in known_pn:
            if (line_id in known_pn[w]):
                w_type = "PN"
        if w in known_gn:
            if (line_id in known_gn[w]):
                if (w in known_pn and w_type != "-"):
                    print("Warning, replacing PN with GN!")
                w_type = "GN"
                if (norm_geo):
                    w = "geoname"
        if professions.replaceProfessions(w) == 'profession':
            if (w_type != "-"):
                print ("Warning, replacing name with profession!")
            w_type = "PF"

        if 'number' in w:
            if (w_type != "-"):
                print ("Warning, overwriting a known type with number: ", w, " - ", w_type)
            w_type = "N"
            w = 'number'
        if last_word == "iti" and norm_date:
            if (w_type != "-"):
                print ("Warning, overwriting a known type with date: ", w, " - ", w_type)
            w_type = "D"
            w = 'date'

        if not (i == 0 and (w == "" or w == "-")):
            if (i > 0 and i < len(words) -1):
                # write last word
                x_index += writeWord(last_tablet, last_line_id, last_index, last_word_2,
                           last_word, w, x_index, out_key, out_features, out_target,
                                     last_name, last_geo, test_run, config)
            elif i == len(words) -1:
                #write last word
                x_index += writeWord(last_tablet, last_line_id, last_index, last_word_2,
                           last_word, w, x_index, out_key, out_features, out_target,
                                     last_name, last_geo, test_run, config)
                #write current word
                x_index += writeWord(tablet_id, line_id, i, last_word, w, "", x_index,
                           out_key, out_features, out_target,
                                     w_type == "PN", w_type == "GN", test_run, config)

        last_word_2 = last_word
        last_word = w
        last_line_id = line_id
        last_tablet = tablet_id
        last_index = i

        if w_type == "PN":
            last_name = True
        else:
            last_name = False
        if w_type == "GN":
            last_geo = True
        else:
            last_geo = False
    # Return the final index.
    return x_index


def writeWord(tablet_id, line_id, i, last_word, word, next_word, x_index,
              out_key, out_features, out_target, PN, GN, test_run, config):
    if len(word) < 1:
        return 0

    if uniqueNames:
        if not test_run and PN and word in test_names:
            return 0
        if test_run and PN and word in dev_names:
            return 0
        if test_run and PN and word:
            test_names.add(word)
        else:
            dev_names.add(word)

    out_key.write("{0}, {1}, {2}, {3}\n".format(tablet_id, line_id, i, word))
    writeSparse(config, out_features, last_word, word, next_word, x_index)
    writeTarget(out_target, PN, GN)
    return 1


def writeTarget(out_target, isName, isGN):
    if isName:
        out_target.write("1\n")
    elif isGN:
        out_target.write("2\n")
    else:
        out_target.write("0\n")

def writeKey(path):
    out = open(path + 'features.KEY', "w")

    offset = 0

    for i in range(len(left_features)):
        out.write("[L]")
        out.write(left_features[i])
        out.write("\n")
    for i in range(len(right_features)):
        out.write("[R]")
        out.write(right_features[i])
        out.write("\n")
    for i in range(len(spelling_features)):
        out.write("[S]")
        out.write(spelling_features[i])
        out.write("\n")

    out.write("-word length-\n")

    for i in range(len(known_symbols)):
        out.write(known_symbols[i])
        out.write("\n")



#!/usr/bin/python3

# Mike Canoy
# Authored     5/3/2016
# Last updated 2/25/2017

import editdistance

left_rules = {}
right_rules = {}
varients = {}
rules = {}               # {rules : {namesperrule : occurrence}}


def main(text, name):
    """
    Args:
        text = line from corpus split into array split on ' '
        name = the PN found in this line

    Returns:
        None (?)
        Probably fills left_rules and right_rules dicts

    Raises:
        None

    """

    text = text.split(' ')
    edit_distance = []
    for word in text:
        edit_distance.append(editdistance.eval(name, word))

    index = edit_distance.index(min(edit_distance))
    varient = text[index]
    if name != varient:
        if name in varients.keys():
            if varient not in varients[name]:
                varients[name].append(varient)
        else:
            varients[name] = [varient]

    if index - 1 >= 0:
        left = text[index - 1]
        if left in left_rules.keys():
            left_rules[left] += 1
        else:
            left_rules[left] = 1

        if left not in rules:
            rules[left] = {name: 1}
        elif left in rules and name not in rules[left].keys():
            rules[left].update({name: 1})
        elif left in rules and name in rules[left].keys():
            rules[left][name] += 1

    if index + 1 <= len(text) - 1:
        right = text[index + 1]
        if right in right_rules.keys():
            right_rules[right] += 1
        else:
            right_rules[right] = 1

        if right not in rules:
            rules[right] = {name: 1}
        elif right in rules and name not in rules[right].keys():
            rules[right].update({name: 1})
        elif right in rules and name in rules[right].keys():
            rules[right][name] += 1



from sner.scripts import professions, utilities
import codecs
import csv
import re
import argparse
import random
from sner.classes import Display


left_features =  []

right_features = []

spelling_features = []

known_symbols = []

x_index = 0

def writeSparse(out_features, word_left, word_middle, word_right, x_index):
    """
    Writes a single x vector of features in a one hot inspired representation
      to the out_features file.

    Args:
      out_features = output file to write features
      word_left = the word left of the word to be output
      word_middle = the word in question
      word_right = the right context of the word to be output
      x_index = the row id for the feature entry

    Returns:
      Nothing

    Raises:
      None

    """

    offset = 0
    for i in range(len(left_features)):
        left = left_features[i]
        if word_left == left:
            out_features.write("{} {} 1\n".format(x_index, i + offset))

    offset = len(left_features)
    for i in range(len(right_features)):
        right = right_features[i]

        if word_right == right:
            out_features.write("{} {} 1\n".format(x_index, i + offset))

    offset = len(left_features) + len(right_features)
    for i in range(len(spelling_features)):
        spelling = spelling_features[i]
        if spelling in word_middle:
            out_features.write("{} {} 1\n".format(x_index, i + offset))

    offset = len(left_features) + len(right_features) + len(spelling_features)
    symbols = word_middle.split('-')
    out_features.write("{} {} {}\n".format(x_index, offset, len(symbols)))
    offset += 1
    foundSym = []
    for sym in symbols:
        if sym in known_symbols and not sym in foundSym:
            foundSym.append(sym)
            out_features.write("{} {} 1\n".format(x_index, known_symbols.index(sym) + offset))



def main(config):
    """
    Finds all names in options.attestations file, and goes through each word
      in the main options.corpus file to create a sparse matrix file to be used
      with scikit learn.

    Args:
        options.norm_num = True to normalize numbers
        options.norm_prof = True to normalize professions
        options.norm_geo = True to normalize geographical names

    Returns:
        .csv file with the format:
            TabletID, LineID, LocationInSentence, Word, Word Type
        Word Types: '-' for unknowns.
                    'PN' to identify personal names.
                    'GN' to identify geographical names.
                    'PF' to identify professions.

    Raises:
        None

    """

    path = config['path']
    corpus = config['corpus']

    features_file = open(path + 'features.KEY', 'r')
    features = features_file.read().splitlines()

    i = 0
    while i < len(features):
        while features[i][:3] == '[L]':
            left_features.append(features[i][3:])
            i += 1
        while features[i][:3] == '[R]':
            right_features.append(features[i][3:])
            i += 1
        while features[i][:3] == '[S]':
            spelling_features.append(features[i][3:])
            i += 1

        if features[i] != '-word length-':
            known_symbols.append(features[i])
            i += 1
        else:
            i += 1


    global x_index
    known_pn = {}
    known_gn = {}

    file2 = codecs.open(path + corpus, 'r', encoding = 'utf-8')

    lines = file2.read().splitlines()

    out_features = open(path + 'features_atf.sparseX', "w")
    out_target = open(path + 'target_atf.RT', "w")
    out_key = open(path + 'target_atf.KEY', "w")
    x_index = 0

    tags = [
        '&',
        '#',
        '@',
        '$',
        '>'
    ]

    display = Display()
    display.start('Writing sparse matrix')

    i = 0
    tablet_id = 0
    while i < len(lines):
        line = lines[i]
        if "@tablet" in line:
            tablet_id += 1
        if not (len(line) > 0 and line[0] in tags):
            j = 0
            while (j < len(line) and line[j].isdigit()):
                j += 1

            if line[j:j + 2] == ". " or line[j:j + 3] == "'. ":
                k = line.index(' ')
                line = line[k + 1:]

                writeLine(
                    config,
                    line,
                    i,
                    out_features,
                    out_target,
                    out_key,
                    known_pn,
                    known_gn,
                    tablet_id
                )
        i += 1
        display.update_progress_bar(i, len(lines))


    display.finish()
    out_features.close()
    out_target.close()


def writeLine(config, line, line_id, out_features, out_target, out_key, known_pn, known_gn, tablet_id):
    norm_date = config['norm']['date']
    norm_geo  = config['norm']['geo']
    norm_num  = config['norm']['num']
    norm_prof = config['norm']['prof']

    global x_index

    # Clean up the line and ensure that there are no extra spaces
    text = utilities.clean_line(line.rstrip(), norm_num, norm_prof)
    text = text.lower()
    text = " ".join(text.split())
    text = text.strip()

    newText = text;
    words = text.split(' ')
    # change to line_id
    last_word = ""
    last_word_2 = ""
    last_line_id = ""
    last_tablet = ""
    last_index = -1

    last_name = False
    last_geo = False

    for i in range(len(words)):
        w = words[i]
        w_type = "-"
        # Set types if they are known and avoid
        # conflicts by using the line_id.
        if norm_prof and professions.replaceProfessions(w) == 'profession':
            if (w_type != "-"):
                print ("Warning, replacing name with profession!")
            w_type = "PF"

        if norm_num and 'number' in w:
            if (w_type != "-"):
                print ("Warning, overwriting a known type with number: ", w, " - ", w_type)
            w_type = "N"
            w = 'number'
        if norm_date and last_word == "iti":
            if (w_type != "-"):
                print ("Warning, overwriting a known type with date: ", w, " - ", w_type)
            w_type = "D"
            if (norm_date):
                w = 'date'

        if not (i == 0 and (w == "" or w == "-" or w == "...")):
            if (i > 0 and i < len(words) -1):
                # write last word
                out_key.write("{0}, {1}, {2}, {3}\n".format(tablet_id, last_line_id, last_index, last_word))
                writeSparse(out_features, last_word_2, last_word, w, x_index)
                x_index += 1
            elif i == len(words) -1:
                #write last word
                out_key.write("{0}, {1}, {2}, {3}\n".format(tablet_id, last_line_id, last_index, last_word))
                writeSparse(out_features, last_word_2, last_word, w, x_index)
                x_index += 1
                #write current word
                out_key.write("{0}, {1}, {2}, {3}\n".format(tablet_id, line_id, i, w))
                writeSparse(out_features, last_word, w, "", x_index)
                x_index += 1

        last_word_2 = last_word
        last_word = w
        last_line_id = line_id
        last_index = i



"""Launcher for scikit-learn models"""
import time
import numpy as np
import pandas as pd
from sklearn import svm, tree
from sklearn.ensemble import RandomForestClassifier, ExtraTreesClassifier
from sklearn.linear_model import SGDClassifier
from sklearn.naive_bayes import MultinomialNB
from sklearn.metrics import accuracy_score, mean_squared_error
from sklearn.metrics import median_absolute_error, precision_score, recall_score
from scipy.sparse import coo_matrix
import random
import os.path

def dec_model(params):
    """Initiates model as decision tree classifier.

    Args:
        params (dict): Dictionary of hyperparameters.

    Returns:
        model (sklearn.tree.DecisionTreeClassifer): Selected model.

    Raises:
        None
    """

    if (params['random']):
        print("Random Decision Tree Parameters.")
        params['criterion'] = random.choice(["gini", "entropy"])
        params['splitter'] = random.choice(["best", "random"])
        params['max_features'] = random.choice(['auto', 'sqrt', 'log2', random.randrange(50, 1000, step=25), None])
        params['max_depth'] =  random.choice([None, random.randrange(5, 1000, step=5)])
        params['min_samples_split'] = random.choice([2, random.randrange(1, 50, step=1)])
        params['max_leaf_nodes'] = random.choice([None, random.randrange(2, 50, step=1)])
        params['min_samples_leaf'] = random.choice([1, random.randrange(5, 100, step=5)])
        print(params)
        
    model = tree.DecisionTreeClassifier(
        criterion=params['criterion'],
        splitter=params['splitter'],
        max_features=params['max_features'],
        max_depth=params['max_depth'],
        min_samples_split=params['min_samples_split'],
        max_leaf_nodes=params['max_leaf_nodes'],
        min_samples_leaf=params['min_samples_leaf']
    )

    return model



def forest_model(params):
    """Initiates model as Extra Trees Classifier.

    Args:
        params (dict): Dictionary of hyperparameters.

    Returns:
        model (sklearn.naive_bayes.MultinomialNB): Selected model.

    Raises:
        None
    """
    if (params['random']):
        params['n_estimators'] = random.choice([1, 3, 5, 10, 20, 30, 40, 50, 75, 100])
    model = ExtraTreesClassifier(
        n_estimators=params['n_estimators'],
        random_state=0
    )

    return model



def nbc_model(params):
    """Initiates model as multinomial Naive Bayes classifier.

    Args:
        params (dict): Dictionary of hyperparameters.

    Returns:
        model (sklearn.naive_bayes.MultinomialNB): Selected model.

    Raises:
        None
    """
    if (params['random']):
        params['alpha'] = random.randrange(1, 10, step=1) * 0.1
    model = MultinomialNB(
        alpha=params['alpha']
    )

    return model


def rdf_model(params):
    """Initiates model as random forest classifier.

    Args:
        params (dict): Dictionary of hyperparameters.

    Returns:
        model (sklearn.ensemble.RandomForestClassifer): Selected model.

    Raises:
        None
    """

    if (params['random']):
        print("Random Decision Tree Parameters.")
        params['criterion'] = random.choice(["gini", "entropy"])
        params['max_features'] = random.choice(['auto', 'sqrt', 'log2', random.randrange(50, 1000, step=25), None])
        params['max_depth'] =  random.choice([None, random.randrange(5, 1000, step=5)])
        params['min_samples_split'] = random.choice([2, random.randrange(1, 50, step=1)])
        params['max_leaf_nodes'] = random.choice([None, random.randrange(2, 50, step=1)])
        params['min_samples_leaf'] = random.choice([1, random.randrange(5, 100, step=5)])
        params['n_estimators'] = random.choice([1, 3, 5, 10, 20, 30, 40, 50, 75, 100])
        print(params)
        
    model = RandomForestClassifier(
        criterion=params['criterion'],
        max_features=params['max_features'],
        max_depth=params['max_depth'],
        min_samples_split=params['min_samples_split'],
        max_leaf_nodes=params['max_leaf_nodes'],
        min_samples_leaf=params['min_samples_leaf'],
        n_estimators=params['n_estimators'],
        verbose=0
    )

    return model


def sgd_model(params):
    """Initiates model as linear classifier with stochastic gradient descent.

    Args:
        params (dict): Dictionary of hyperparameters.

    Returns:
        model (sklearn.linear_model.SGDClassifer): Selected model.

    Raises:
        None
    """

    
    if (params['random']):
        params['loss'] = random.choice(['hinge', 'log', 'modified_huber', 'squared_hinge', 'perceptron', 'squared_loss', 'huber', 'epsilon_insensitive', 'squared_epsilon_insensitive'])
        params['penalty'] = random.choice(['none', 'l2', 'l1', 'elasticnet'])
        params['alpha'] = random.choice([0.001, 0.0001, 0.00001])
    model = SGDClassifier(
        loss=params['loss'],
        penalty=params['penalty'],
        alpha=params['alpha']
    )

    return model


def svc_model(params):
    """Initiates model as c-support vector classifier.

    Args:
        params (dict): Dictionary of hyperparameters.

    Returns:
        model (sklearn.svm.SVC): Selected model.

    Raises:
        None
    """

    if (params['random']):
        params['C'] = random.randrange(1, 13, step=1) * 0.1
        params['degree'] = random.choice([2, 3, 4, 5])
        params['kernel'] = random.choice(['linear', 'rbf', 'poly', 'sigmoid'])
        
    model = svm.SVC(
        decision_function_shape='ovo',
        C=params['C'],
        degree=params['degree'],
        kernel=params['kernel']
    )

    return model


def main(config):
    """Load and train a scikit-learn model as specified by the user in configs
    or command line arguments.  Test the models performance after training and
    then print performance information (precision, accuracy, recall).

    Args:
        config (dict): Configuration flags and values.

    Returns:
        None

    Raises:
        None
    """

    runModel(config)
    while (config['looping']):
        runModel(config)

def runModel(config):

    model_name = config['run']
    params = config['params']
    path = config['path']
    use_atf = config['flags']['use_atf']
    use_test = config['eval-test']

    if use_atf:
        print("Exporting ATF prediction.")
    
    output_path = path + "{}_{}_training_results.csv".format(
        model_name,
        time.strftime('%Y%m%d_%H%M')
    )


	
    #read the input data per the configuration information

    if (config['output-file'] != None):
        output_path = config['output-file']

    print("Reading data")
    feature_count = sum(1 for line in open(path + 'features.KEY', 'r'))

    Y = np.loadtxt(path + 'target_train.RT')
    A = np.loadtxt(path + 'features_train.sparseX', ndmin=2)
    I = A[:, 0]
    J = A[:, 1]
    data = A[:, 2]
    X = coo_matrix((data, (I, J)), shape=(int(max(I))+1, feature_count))


    devY = np.loadtxt(path + 'target_dev.RT')
    A = np.loadtxt(path + 'features_dev.sparseX', ndmin=2)
    I = A[:, 0]
    J = A[:, 1]
    data = A[:, 2]
    devX = coo_matrix((data, (I, J)), shape=(int(max(I))+1, feature_count))

    testX = None
    testY = None    
    if use_test:
            testY = np.loadtxt(path + 'target_test.RT')            
            A = np.loadtxt(path + 'features_test.sparseX', ndmin=2)
            I = A[:, 0]
            J = A[:, 1]
            data = A[:, 2]

            testX = coo_matrix((data, (I, J)), shape=(int(max(I))+1, feature_count))
    
    atfX = None
    if use_atf:
        A = np.loadtxt(path + 'features_atf.sparseX', ndmin=2)
        I = A[:, 0]
        J = A[:, 1]
        data = A[:, 2]
        atfX = coo_matrix((data, (I, J)), shape=(int(max(I))+1, feature_count))
	
    print("Training model")

	#select the model based on the configuration data
    if model_name == 'dec':
        model = dec_model(params)
    elif model_name == 'nbc':
        model = nbc_model(params)
    elif model_name == 'rdf':
        model = rdf_model(params)
    elif model_name == 'sgd':
        model = sgd_model(params)
    elif model_name == 'svc':
        model = svc_model(params)
    elif model_name == 'forest':
        model = forest_model(params)

	#train the model on the annotated data
    model.fit(X, Y)

    prediction = model.predict(X)
    trainMSE = mean_squared_error(Y, prediction)
    print("Train Error: {:.3f}".format(trainMSE))

	#generate predictions based off of the model, and evaluate its performance
    prediction = model.predict(devX)
    devMSE = mean_squared_error(devY, prediction)
    devMAE = median_absolute_error(devY, prediction)
    score = model.score(devX, devY)
    acc = accuracy_score(devY, prediction)
    prec = precision_score(devY, prediction, average=None)
    recall = recall_score(devY, prediction, average=None)
    avg_prec = (sum(prec) / len(prec))
    avg_recall = (sum(recall) / len(recall))
    print("Acc: %.5f Prec: %.4f Recall: %.4f"%(acc, avg_prec, avg_recall))
    #print("Precision: {:.4f}".format(avg_prec))
    #print(prec)
    #print("Recall: {:.4f}".format(avg_recall))
    #print(recall)

	#output results, if use_atf is not set then output a .csv file
    if not use_atf:
        data = params
        data.update({
        'avg_prec': avg_prec,
            'prec[0]': prec[0],
            'prec[1]': prec[1],
            'prec[2]': prec[2],
            'avg_recall': avg_recall,
            'recall[0]': recall[0],
            'recall[1]': recall[1],
            'recall[2]': recall[2]
        })

        useHeader = True
        if (os.path.isfile(output_path)):
            useHeader = False
        
        with open(output_path, 'a') as f:
            output = pd.DataFrame(data, index=[1])
            output.to_csv(f, header=useHeader)

    np.savetxt(
        path + 'dev_prediction.RT',
        prediction,
        delimiter=',',
        fmt='%d'
    )


    #if use_test is set, then evaluate test data as well
    if use_test:
        print("Evaluating on test dataset...")
        prediction = model.predict(testX)
        testMSE = mean_squared_error(testY, prediction)
        testMAE = median_absolute_error(testY, prediction)
        score = model.score(testX, testY)
        acc = accuracy_score(testY, prediction)
        prec = precision_score(testY, prediction, average=None)
        recall = recall_score(testY, prediction, average=None)
        avg_prec = (sum(prec) / len(prec))
        avg_recall = (sum(recall) / len(recall))
        print("Score: %.5f MSE: %.5f MAE: %.5f "% (score, testMSE, testMAE))
        print("Acc: %.5f Prec: %.4f Recall: %.4f"%(acc, avg_prec, avg_recall))
        print("Precision: %.5f, %.5f, %.5f (NotName, PN, GN)"%(prec[0], prec[1], prec[2]))
        print("Recall: %.5f, %.5f, %.5f (NotName, PN, GN)"%(recall[0], recall[1], recall[2]))

    #if use_atf is set, then output a .atf file containing the results        
    if use_atf:
        print("Predicting...")
        atf_prediction = model.predict(atfX)
        print("Saving...")
        np.savetxt(
            path + 'atf_prediction.RT',
            atf_prediction,
            delimiter=',',
            fmt='%d'
        )


    #export image of the tree if the decision tree model was used and not using random params
    if model_name == 'dec' and params['random'] == False:
        with open(path + 'decModel.dot', 'w') as f:
            with open(path + 'features.KEY', 'r') as f2:
                feature_labels = f2.read().splitlines()
                class_labels = (
                    'NotName',
                    'PersonalName',
                    'GeoName'
                )
                f = tree.export_graphviz(
                    model,
                    out_file=f,
                    feature_names=feature_labels,
                    class_names=class_labels
                )

    if model_name == 'forest':
        importances = model.feature_importances_
        std = np.std([tree.feature_importances_ for tree in model.estimators_], axis=0)
        indices = np.argsort(importances)[::-1]
        with open('FeatureRankings.txt', 'w') as feats:
            feats.write("Feature ranking:\n")
            with open(path + 'features.KEY', 'r') as f2:
                feature_labels = f2.read().splitlines()
                for f in range(X.shape[1]):
                    feats.write("%d. feature %d (%f) [%s]\n" % (f + 1, indices[f], importances[indices[f]], feature_labels[f]))



from sner.scripts import professions, utilities
import codecs
import csv
import re
import argparse


def findKnown(data, options, knownPN, knownGN):
    """
    Iterates through the attenstations file for personal and geographical names
     and adds the lineIDs in a list to the knownPN and knownGN dictionaries.
    Args:
        data (?)
        options (?)
        knownPN = { Personal name : line ID }, to be filled in.
        knownGN = { Geographical name : line ID}, to be filled in.

    Returns:
        Updates knownPN and knownGN dictionaries with the lineIDs.

    Raises:
        None

    """

    file = codecs.open(data.attestations, 'r', encoding = 'utf-8')
    # find all of the names
    for line in file:
        line = line.split(',')
        text = utilities.clean_line(line[4].rstrip(), options.norm_num, options.norm_prof)
        text = text.lower()
        name = utilities.clean_line(line[5].rstrip(), options.norm_num, options.norm_prof)
        name = name.lower()
        lineID = line[1]
        lineID = re.sub("[L]", "", lineID)

        if line[9].rstrip() == 'PN':
            if (name not in knownPN):
                knownPN[name] = [lineID]
            else :
                knownPN[name].append(lineID)
        if line[9].rstrip() == 'GN':
            if (name not in knownGN):
                knownGN[name] = [lineID]
                if (name in knownPN):
                    print ("Note: ", name, " is both a PN and GN.")
            else :
                knownGN[name].append(lineID)

def main(data, options):
    """
    Finds all names in options.attestations file, and goes through each word
      in the main options.corpus file to fill a new output file specified by
      options.output
    Args:
        options.norm_num = True to normalize numbers
        options.norm_prof = True to normalize professions
        options.norm_geo = True to normalize geographical names

    Returns:
        .csv file with the format:
            TabletID, LineID, LocationInSentence, Word, Word Type
        Word Types: '-' for unknowns.
                    'PN' to identify personal names.
                    'GN' to identify geographical names.
                    'PF' to identify professions.

    Raises:
        None

    """

    knownPN = {}
    knownGN = {}
    findKnown(data, options, knownPN, knownGN)

    # iterate each line in Garshana Text and output to specified file
    file2 = codecs.open(data.corpus, 'r', encoding = 'utf-8')
    out = open(data.output, "w")

    out.write("Tablet ID,Line Number,Word Number,Word,Word Type\n")

    for line in file2:
        line = line.split(',')
        tabletID = line[0]

        # Clean up the line and ensure that there are no extra spaces
        text = utilities.clean_line(line[7].rstrip(), options.norm_num, options.norm_prof)
        text = text.lower()
        text = " ".join(text.split())
        text = text.strip()

        # We should skip the first line that labels the csv
        if "text" in text:
            continue

        newText = text;
        words = text.split(' ')
        lineID = line[6]
        # skip description lines
        if lineID == "":
            continue
        # change to lineID
        lineID = line[4]
        lineID = re.sub("[L]", "", lineID)
        lastWord = ""
        if "'" in lineID:
            lineID = re.sub("[']", "", lineID)
            lineID = "'" + lineID
        for i in range(len(words)):
            w = words[i]
            wType = "-"
            # Set types if they are known and avoid
            # conflicts by using the lineID.
            if w in knownPN:
                if (lineID in knownPN[w]):
                    wType = "PN"
            if w in knownGN:
                if (lineID in knownGN[w]):
                    if (w in knownPN and wType != "-"):
                        print("Warning, replacing PN with GN!")
                    wType = "GN"
                    if (options.norm_geo):
                        w = "geoname"
            if professions.replaceProfessions(w) == 'profession':
                if (wType != "-"):
                    print ("Warning, replacing name with profession!")
                wType = "PF"

            if 'number' in w:
                if (wType != "-"):
                    print ("Warning, overwriting a known type with number: ", w, " - ", wType)
                wType = "N"
                w = 'number'
            if lastWord == "iti":
                if (wType != "-"):
                    print ("Warning, overwriting a known type with date: ", w, " - ", wType)
                wType = "D"
                if (options.norm_date):
                    w = 'date'

            lastWord = w
            if not (i == 0 and (w == "" or w == "-" or w == "...")):
                out.write("{},{},{},{},{}\n".format(tabletID, lineID, i, w, wType))



"""setup.py: From github.com/pypa/sampleproject"""
from setuptools import setup, find_packages


with open('README.rst') as readme:
    long_description = readme.read()

setup(
    name='sumerian-ner',
    version='0.1.1',
    description='Sumerian Named Entity Recognition',
    long_description=long_description,
    url='https://github.com/wwunlp/sner',
    author='WWUNLP SNER Team',
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=[
        'editdistance',
        'matplotlib',
        'numpy',
        'pandas',
        'scikit-learn',
        'scipy'
    ],
    entry_points={
        'console_scripts': [
            'sner=sner:main',
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/wwunlp/sner/issues',
        'Source': 'https://github.com/wwunlp/sner',
    }
)



"""Names from rule."""
from sner.classes import Rule


def main(corpus, rule):
    """
    Takes the overall corpus as input, as well as a single Rule object.
    Will return a Set of tokens satisfying the passed-in Rule, referred to in
    the code as 'names'.

    Args:
        corpus (set): Set of all Token objects in the corpus.
        rule (Rule): Rule object used to identify results.

    Returns:
        names (set): Set of Token objects that match the passed in Rule.

    Raises:
        TypeError

    """

    names = set()

    if rule.type == Rule.Type.left_context:
        for token in corpus:
            if rule.contents == token.left_context:
                names.add(token)
                token.rules.add(rule)

    elif rule.type == Rule.Type.right_context:
        for token in corpus:
            if rule.contents == token.right_context:
                names.add(token)
                token.rules.add(rule)

    elif rule.type == Rule.Type.spelling:
        for token in corpus:
            if rule.contents in token.word:
                names.add(token)
                token.rules.add(rule)

    elif rule.type == Rule.Type.unset:
        raise TypeError("improperly initialized rule within ruleset: " +
                        rule.contents + ", type unset")

    else:
        raise TypeError("unrecognized rule type found in nameFromRule: " +
                        rule.content + ", of type: " + rule.type)

    return names



"""
Generates dictionaries of monograms, bigrams, and trigrams, mapping them to
lists containing the number of times they occur in names, and the number of
times they occur in total.  This is information later used to evaluate these
grams usefulness as name recognition rules.
"""

from sner.scripts import readnames, utilities
from collections import defaultdict

monograms = defaultdict(list)
bigrams = defaultdict(list)
trigrams = defaultdict(list)
rules = {}

def gramhelper(gramDict, gram, namecount, totalcount):
    """
	Utility function to insert arbitrary n-grams into dictionaries that
	associate them with lists containing the number of times that gram occurred
	as well as the number of times that gram occurred in a name.

    Args:
            gramDict = dictionary to recieve an n-gram
            gram = gram to be inserted into the dict
            namecount = the occurrences for this n-gram in a name
            totalcount = how many times this gram has occurred in the corpus in any word

    Returns:
            Updates gramDict with the information from gram

    Raises:

    """
    list = gramDict[gram]

    if(len(list) == 0):
        list.append(0)
        list.append(0)

    list[0] += totalcount
    list[1] += namecount


"""
The following three functions work similarly, adding the gram and count of an
  individual name and total count of that gram occurring to the dictionary of
  the related gram size.
Args:
    gram = string version of gram to be added
    namecount = how many times this particular name has occurred (?)
    totalcount = how many times this gram has occurred in the corpus

Returns:
    Updated gram mapping, increasing the count of that name occurring and that
      individual gram occurring.

Raises:

"""

def addMonogram(gram, namecount, totalcount):
    gramhelper(monograms, gram, namecount, totalcount)


def addBigram(gram, namecount, totalcount):
    gramhelper(bigrams, gram, namecount, totalcount)


def addTrigram(gram, namecount, totalcount):
    gramhelper(trigrams, gram, namecount, totalcount)

# Load name data


def loadData(data, allgrams, f):
    """
    Loads name data
    Args:
        data(string) = name of the corpus data file
        allgrams(dict) = dictionary of all grams of the form:
                   { gram : namecount, totalcount }
        f (file) = object used to access the output file

    Returns:
        Fills allgrams dictionary.

    Raises:
        If grams are found in name dataset but not in overall dataset, will
          print out the type of gram and oddity.

    """

    # Load ngrams found in names
    namegrams = readnames.getKgrams(readnames.getPNs(data), 3)

    if len(namegrams) >= 1:
        print("loading monograms in names")
        for k, v in namegrams[0].items():
            klower = k.lower()
            if klower in allgrams:
                totalcount = allgrams[klower]
                addMonogram(klower, v, totalcount)
            else:
                print("ERROR: monogram \"{0}\" found in name dataset but" +
                      "not overall dataset".format(klower))

    print()

    if len(namegrams) >= 2:
        print("loading bigrams in names")
        for k, v in namegrams[1].items():
            klower = k.lower()
            if klower in allgrams:
                totalcount = allgrams[klower]
                addBigram(klower, v, totalcount)
            else:
                print("ERROR: bigram \"{0}\" found in name dataset but" +
                      "not overall dataset".format(klower))

    print()

    if len(namegrams) >= 3:
        print("loading trigrams in names")
        for k, v in namegrams[2].items():
            klower = k.lower()
            if klower in allgrams:
                totalcount = allgrams[klower]
                addTrigram(klower, v, totalcount)
            else:
                print("ERROR: trigram \"{0}\" found in name dataset but" +
                      "not overall dataset".format(klower))

# Collect percentage statistics from the gram maps


def analyzeData(f):
    """
    Collect percentage statistics from the gram maps and writes them to a .csv
	output file.
    Args:
        f(file) = the csv file in utf-8 format being written to, with fields:
            N-Gram, Percentage, Occurence, Total Occurence

    Returns:
        None

    Raises:
		None

    """

    f.write("N-gram, Percentage, Occurence, Total Occurence\n".encode('utf-8'))

    print('monogram analysis')
    for k, v in monograms.items():
        outputAnalysis(k, v, f)
    if len(monograms) <= 0:
        print("no monograms")
    print("done.")

    print('\nbigram analysis')
    for k, v in bigrams.items():
        outputAnalysis(k, v, f)
    if len(bigrams) <= 0:
        print("no bigrams")
    print("done.")

    print('\ntrigram analysis')
    for k, v in trigrams.items():
        outputAnalysis(k, v, f)
    if len(trigrams) <= 0:
        print("no trigrams")
    print("done.")

    f.close()


def outputAnalysis(k, v, f):
    """
    Args:
        k(string) =
        v = (?)
        f = csv file we are writing to.
    Returns:

    Raises:
        If an n-gram has greater than 100% significance, reports the n-gram.

    """

    significance = v[1] / v[0]
    if significance > 1:
        print("ERROR: ngram \"{0}\" has greater than 100% significance" +
              "({1: .4}) {2}: {3}".format(k, significance, v[1], v[0]))
    f.write("{0},{1},{2},{3}\n".format(k,
            significance, v[1], v[0]).encode('utf-8'))


# Main


def main(data, syll_count):
    f = open('results/spelling.csv', 'wb')
    loadData(data, syll_count, f)
    print()
    analyzeData(f)



"""NER Model"""
from sner.classes import Display, Rule, Token
from sner.scripts.ner import contextualfromnames, namesfromrule
from sner.scripts.ner import spellingfromnames, updatetokenstrength
import matplotlib.pyplot as plt
import pandas as pd


def import_corpus(corpus_path, display):
    """
	Imports the corpus for the ner model, from a format of our own design.
	The format is a CSV containing individual words in the corpus, with columns
	containing the Tablet ID, the line number within that tablet, the word
	number within that line, the word itself, and any annotation associated
	with that word.

    Args:
        corpus_path (str): Path of the corpus file.
		display (Display): Utility object used to print the progress of scanning
						   the corpus file.

    Returns:
        corpus (set): Set of Token objects, properly initialized from the input
					  data.

    Raises:
        None

    """

    corpus = set()

    cols = ['Tablet ID', 'Line Number', 'Word Number', 'Word', 'Word Type']

    data = pd.read_csv(corpus_path, names=cols, header=0)
    for i in range(len(data) - 1):
        if data.loc[i, 'Word Number'] > 0:
            left_context = data.loc[i - 1, 'Word']
        else:
            left_context = None

        word = data.loc[i, 'Word']

        if data.loc[i, 'Line Number'] == data.loc[i + 1, 'Line Number']:
            right_context = data.loc[i + 1, 'Word']
        else:
            right_context = None

        word_type = Token.find_type(data.loc[i, 'Word Type'])

        token = Token(left_context, word, right_context, word_type)
        corpus.add(token)

        display.update_progress_bar(i + 1, len(data))

    return corpus


def import_seed_rules(seed_rules_path, display):
    """
    This function will read the seed rule file, and return the contents as
    a set.  The file is a CSV formatted with columns containing the rule type,
	the contents of the rule, and its strength rating (the probability that a
	token is a name if the rule applies to said token).

    Args:
        rulename (str): Location of seed rules file.

    Returns:
        rules (set): Set of Rule objects corresponding to the rules in the
            seed rules file.

    Raises:
		None

    """

    seed_rules = set()

    cols = ['Rule Type', 'Rule', 'Strength']

    data = pd.read_csv(seed_rules_path, names=cols, header=0)
    for i in range(len(data) - 1):
        rule_type = Rule.find_type(data.loc[i, 'Rule Type'])
        rule = data.loc[i, 'Rule']
        strength = data.loc[i, 'Strength']
        rule = Rule(rule_type, rule, strength)
        seed_rules.add(rule)

        display.update_progress_bar(i + 1, len(data))

    return seed_rules


def assess_strength(rules, corpus, config):
    """
	Evaluates the accuracy of the strength rating of the passed-in rules.  This
	is useful because the ner model will generate rules in an unsupervised
	fashion.  This function gets used to evaluate the performance of that
	process.

    Args:
        rules (set): A set of Rule objects to be evaluated
        corpus (set): A set of Token objects, representing the entire Garshana
					  corpus.

    Returns:
        None

    Raises:
        None

    """

    bad_rules = 0
    bad_context = 0
    bad_spelling = 0

    total_context = 0
    total_spelling = 0
    total_delta = 0

    est_false_positives = 0

    print("rule performance:")
    print("calculating...", end='\r')

    i = 0
    cols = {
        'Iteration'    : [],
        'Rule'         : [],
        'Type'         : [],
        'Strength'     : [],
        'True Strength': [],
        'Occurrences'  : []
    }
    output = pd.DataFrame(data=cols)

    x_vals = []
    y_vals = []
    rule_num = 1

    for rule in rules:
        names = namesfromrule.main(corpus, rule)
        real_names = 0
        total_names = len(names)

        for token in names:
            if token.type == Token.Type.personal_name:
                real_names += 1

        if total_names == 0:
            true_strength = 0
        else:
            true_strength = real_names / total_names

        delta = abs(true_strength - rule.strength)
        total_delta += delta

        x_vals.append(rule_num)
        rule_num += 1
        y_vals.append(delta)

        if rule.type == Rule.Type.spelling:
            total_spelling += 1
        else:
            total_context += 1

        #if a rule is more than 20% from its true value, it is 'bad'
        if delta > 0.2:
            bad_rules += 1
            if rule.type == Rule.Type.spelling:
                bad_spelling += 1
            else:
                bad_context += 1

        i += 1
        output.loc[i, 'Iteration']     = rule.iteration
        output.loc[i, 'Rule']          = rule.contents
        output.loc[i, 'Type']          = rule.type.name
        output.loc[i, 'Strength']      = rule.strength
        output.loc[i, 'True Strength'] = true_strength
        output.loc[i, 'Occurrences']   = rule.occurrences

    output_path = config['path'].format(
        'ner',
        time.strftime('%Y%m%d_%H%M'),
        'output.csv'
    )
    output.to_csv(path_or_buf=output_path)

    print("               ", end='\r')
    print("percentage of bad rules:    {}%".format(
        100 * bad_rules / len(rules)
    ))
    print("percentage of bad context:  {}%".format(
        100 * bad_context / total_context
    ))
    print("percentage of bad spelling: {}%".format(
        100 * bad_spelling / total_spelling
    ))
    print("average delta value:        {}%".format(
        100 * total_delta / len(rules)
    ))

    plt.xlabel('Rules')
    plt.ylabel('Delta')
    plt.title('Plot of Delta per Rule')
    plt.plot(x_vals, y_vals, 'ro')
    plt.axis([min(x_vals), max(x_vals), min(y_vals), max(y_vals)])
    plt.show()

    sort_y = sorted(y_vals)
    plt.xlabel('')
    plt.ylabel('Delta')
    plt.title('Delta Sorted')
    plt.plot(x_vals, sort_y, 'ro')
    plt.axis([min(x_vals), max(x_vals), min(sort_y), max(sort_y)])
    plt.show()


def get_new_names(corpus, names, rules):
    """
    Meant to use the provided ruleset to scan the corpus for new names.
    It will then return the names in quesiton, which will be used
    to generate more rules.

	Basically, it grabs all tokens from the corpus matching the rules in
	question and then return them as a set.  The names parameter lets you
	specify tokens that are already recognized as names, allowing you to
	retrieve only new name results.

    Args:
        corpus (set): Set of Token objects representing the entire Garshana
					  corpus.
        names (set): Set of Tokens already recognized as names.
        rules (set): Set of Rule objects used to find new names

    Returns:
        new_names (set): Set of Token objects

    Raises:
        None

    """

    new_names = set()

    for rule in rules:
        results = namesfromrule.main(corpus, rule)
        for name in results:
            if name not in names:
                new_names.add(name)

    return new_names


def print_precision_and_recall(selected_elements, relevant_elements, i, log):
    """
	Prints the precision, recall, and F1 score of the algorithm.  Used by
	passing in tokens in the selected_elements parameter.  These tokens are
	the tokens considered to be names.  Relevant elements is just the total
	number of names in the corpus.

    Args:
        selected_elements (set): Set of Token objects representing names as
								 identified by the algorithm.
        relevant_elements (int): Total number of names that exist in the
								 corpus.
        i (int): Index of current log entry.
		log (pandas.DataFrame): Data structure containing logs

    Returns:
		None

    Raises:
        None

    """

    positives = 0.0
    true_positives = 0.0
    precision = 0.0
    recall = 0.0
    f1_score = 0.0

    for token in selected_elements:
        positives += 1.0
        if token.type == Token.Type.personal_name:
            true_positives += 1.0

    if positives == 0.0:
        precision = 0.0
    else:
        precision = true_positives / positives * 100

    if relevant_elements == 0.0:
        recall = 0.0
    else:
        recall = true_positives / relevant_elements * 100

    if precision + recall == 0.0:
        f1_score = 0.0
    else:
        f1_score = 2.0 * precision * recall / (precision + recall)

    print("Precision: {:06.4f}%".format(precision))
    print("Recall:    {:06.4f}%".format(recall))
    print("F1 Score:  {:06.4f}\n".format(f1_score))

    log.loc[i, 'Precision'] = precision
    log.loc[i, 'Recall'] = recall
    log.loc[i, 'F1 Score'] = f1_score


def main(config):
    """
    Rules and names will be lists of RuleSets or TokenSets.
    These sets will represent the results of various iterations of
    the algorithm. So index 0 of rules would be the first rule set
    (seed rules) and 1 would be the first rules generated and used by the
    algorithm itself. Index zero of names would be the names that came from
    the seed rules. Index one the rules that came from rule set 1. And so on.

    Args:
        config (dict): Dictionary containing confiruation information, such as
					   the location of the input files, as well as various
					   flags and runtime parameters.  (defined in sner.py)

    Returns:
        None

    Raises:
        None

    """

    path = config['path']
    corpus_path = path + config['corpus']
    seed_rules_path = path + config['seed-rules']
    iterations = config['iterations']
    max_rules = config['max_rules']

    log_cols = {
        'Iteration Type': [],
        'New Rules'     : [],
        'New Names'     : [],
        'Precision'     : [],
        'Recall'        : [],
        'F1 Score'      : []
    }
    log = pd.DataFrame(data=log_cols)

    display = Display()

    display.start("Importing Corpus: {}".format(corpus_path))
    corpus = import_corpus(corpus_path, display)
    display.finish()

    display.start("Importing Seed Rules: {}".format(seed_rules_path))
    seed_rules = import_seed_rules(seed_rules_path, display)
    display.finish()

    updatetokenstrength.main(corpus, seed_rules)

    rule_set = set()
    rule_set = rule_set.union(seed_rules)

    name_set = set()
    new_names = get_new_names(corpus, name_set, seed_rules)
    name_set = name_set.union(new_names)

    relevant_elements = 0.0
    for token in corpus:
        if token.type == Token.Type.personal_name:
            relevant_elements += 1.0

    display.start(
        "Starting Model:\n   {} iterations, {} max rules".format(
            iterations,
            max_rules
        )
    )
    for i in range(1, iterations + 1):
        if i % 2 == 0:
            iter_type = 'context'
            get_new_rules = contextualfromnames.main
        else:
            iter_type = 'spelling'
            get_new_rules = spellingfromnames.main

        print("Iteration {}: find {} rules".format(i, iter_type))

        new_rules = get_new_rules(
            corpus,
            rule_set,
            name_set,
            max_rules,
            i,
            config,
            display
        )

        print("Found {} new {} rules".format(len(new_rules), iter_type))

        rule_set = rule_set.union(new_rules)
        updatetokenstrength.main(corpus, rule_set)

        new_names = get_new_names(corpus, name_set, new_rules)
        name_set = name_set.union(new_names)

        print("top {} rules found {} new names".format(
            len(new_rules), len(new_names)))

        log.loc[i, 'Iteration Type'] = iter_type
        log.loc[i, 'New Rules'] = len(new_rules)
        log.loc[i, 'New Names'] = len(new_names)
        print_precision_and_recall(name_set, relevant_elements, i, log)

    display.finish()
    log.to_csv(path_or_buf=data.log)
    assess_strength(rule_set, corpus, data)




import re

professions = ["aga'us",
               "arad",
               "aszgab",
               "azlag",
               "ama-gal",
               "bahar",
               "bisajdubak",
               "damgar",
               "dikud",
               "dubsar",
               "en",
               "er2",
               "ereszdijir",
               "ensik",
               "engar",
               "enkud",
               "gaba'asz",
               "galamah",
               "gala",
               "geme",
               "gudug",
               "guzala"
               "ha-za-num2",
               "idu",
               "iszib",
               "kaguruk",
               "kasz",
               "lu2-kasz",
               "ka-us2-sa",
               "kijgia",
               "kinkin",
               "kur-gara",
               "kuruszda",
               "kusz",
               "lu2-mar-sa-me",
               "lu2-ur3-ra",
               "lugal",
               "lukur",
               "lungak",
               "malah",
               "maszkim",
               "muhaldim",
               "muszendu",
               "nagada",
               "nagar",
               "nar",
               "nin",
               "nubanda",
               "nukirik",
               "sajDUN",
               "sajja",
               "simug",
               "sipad",
               "sukkal",
               "szabra",
               "szagia",
               "szakkanak",
               "ša-ra-ab-du",
               "szej",
               "szesz",
               "szidim",
               "szu'i",
               "szukud",
               "szusz3",
               "tibira",
               "ugula",
               "ugula geš2-da",
               "unud",
               "urin",
               "ujjaja",
               "uszbar",
               "zabardab",
               "zadim"]

prof_re = re.compile(r'\b(?<!-)' + r'(?![\w-])|\b(?<!-)'.join(professions) +
                     r'(?![\w-])')


def replaceProfessions(line):
    """
    Replaces known professions with 'profession'
    """
    line, numReplaces = re.subn(prof_re, 'profession', line)
    return line


def main():
    """
    Testing function
    """
    line = 'testing 1234 zadim abc df3urin33  unud 3223aa ne33ds33 345'
    line = re.sub(r'\b(?<!-)(\d+)(?![\w-])', 'number', line)
    line = replaceProfessions(line)
    print(line)

if __name__ == '__main__':
    main()
